"""Copyright © 2023 Pixelgen Technologies AB."""

from .adapterqc import BaseAdapterQCTestsMixin
from .amplicon import BaseAmpliconTestsMixin
from .analysis import BaseAnalysisTestsMixin
from .annotate import BaseAnnotateTestsMixin
from .collapse import BaseCollapseTestsMixin
from .config import WorkflowConfig
from .demux import BaseDemuxTestsMixin
from .graph import BaseGraphTestsMixin
from .layout import BaseLayoutTestsMixin
from .preqc import BasePreQCTestsMixin
from .report import BaseReportTestsMixin
from .workflow import PixelatorWorkflowTest
from .workflow_context import PixelatorWorkflowContext, use_workflow_context

# isort: off
# Unsorted import to avoid circular dependencies
from .collector import YamlIntegrationTestsCollector
# isort: on

__all__ = [
    "BaseAmpliconTestsMixin",
    "BasePreQCTestsMixin",
    "BaseAdapterQCTestsMixin",
    "BaseDemuxTestsMixin",
    "BaseCollapseTestsMixin",
    "BaseGraphTestsMixin",
    "BaseAnnotateTestsMixin",
    "BaseLayoutTestsMixin",
    "BaseAnalysisTestsMixin",
    "BaseReportTestsMixin",
    "PixelatorWorkflowContext",
    "PixelatorWorkflowTest",
    "use_workflow_context",
    "WorkflowConfig",
    "YamlIntegrationTestsCollector",
]



"""Implementation of the Ramer–Douglas–Peucker line simplification algorithm.

Based on https://github.com/fhirschmann/rdp, see license statement below.

Copyright © 2024 Pixelgen Technologies AB.
Copyright (c) 2014 Fabian Hirschmann <fabian@hirschmann.email>

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import numpy as np
import numpy.typing as npt


def _pldist(
    point: np.ndarray, start: np.ndarray, end: np.ndarray
) -> npt.NDArray[np.float64]:
    """Calculate the distance of points to a line segment.

    The line segment is defined by two points, start and end.
    The distance is calculated for each point in the points array.

    :param point: a point
    :type point: numpy array
    :param start: a point of the line
    :type start: numpy array
    :param end: another point of the line
    :type end: numpy array
    """
    if np.all(start == end):
        return np.linalg.norm(point - start)  # type: ignore

    # normalized tangent vector
    d = np.divide(end - start, np.linalg.norm(end - start))

    # signed parallel distance components
    s = np.dot(start - point, d)
    t = np.dot(point - end, d)

    # clamped parallel distance
    h = np.max([s, t, np.zeros_like(s)])

    # perpendicular distance component, as before
    c = np.cross(point - start, d)

    # use hypot for Pythagoras to improve accuracy
    return np.hypot(h, c)


def _ramer_douglas_peucker_iterative(
    M: np.ndarray, start_index: int, last_index: int, epsilon: float
) -> np.ndarray:
    # Initialize a stack to keep track of the ranges that need to be processed
    stack = []
    stack.append([start_index, last_index])

    # Create mask to mark wich points to keep or remove
    indices = np.ones(last_index - start_index + 1, dtype=bool)

    while stack:
        start_index, last_index = stack.pop()

        # Skip empty ranges
        if start_index + 1 == last_index:
            continue

        start, end = M[start_index], M[last_index]

        index_range = np.arange(start_index + 1, last_index)
        index_range_mask = np.nonzero(indices[start_index + 1 : last_index])[0]
        masked_index_range = index_range[index_range_mask]

        # Find the distances of each point to the line segment start <-> end
        dists = _pldist(M[masked_index_range], start, end)

        # map local index for a subrange of points back to the global index
        local_max_index = np.argmax(dists)
        max_distance = dists[local_max_index]
        max_index = masked_index_range[local_max_index]

        # The point is further than epsilon to the line segment
        # split the range and add the new ranges to the stack
        if max_distance > epsilon:
            stack.append([start_index, max_index])
            stack.append([max_index, last_index])

        # The point is closer than epsilon to the line segment, remove it
        else:
            indices[start_index + 1 : last_index] = False

    return indices


def simplify_line_rdp(
    coordinates: np.ndarray, epsilon: float = 1e-2, return_mask=False
):
    """Simplifies a given array of points using the Ramer-Douglas-Peucker algorithm.

    .. example:

    >>> simplify_line_rdp(np.array([[1, 1], [2, 2], [3, 3], [4, 4]]))
    [[1, 1], [4, 4]]

    :param M: a series of points
    :type M: numpy array with shape ``(n,d)`` where ``n`` is the number of points and ``d`` their dimension
    :param epsilon: epsilon in the rdp algorithm
    :type epsilon: float
    :param dist: distance function
    :type dist: function with signature ``f(point, start, end)`` -- see :func:`rdp.pldist`
    :param algo: either ``iter`` for an iterative algorithm or ``rec`` for a recursive algorithm
    :type algo: string
    :param return_mask: return mask instead of simplified array
    :type return_mask: bool
    """
    mask = _ramer_douglas_peucker_iterative(
        coordinates, 0, len(coordinates) - 1, epsilon
    )

    if return_mask:
        return mask

    return coordinates[mask]



"""Filter classes for the PNA assay pipeline.

Copyright © 2024 Pixelgen Technologies AB.
"""

from cutadapt.info import ModificationInfo
from cutadapt.predicates import Predicate

from pixelator.pna.config import PNAAssay, get_position_in_parent


class TooManyN(Predicate):
    """Select reads that have too many 'N' bases.

    Both a raw count or a proportion (relative to the sequence length) can be used.
    """

    def __init__(self, count: float, assay: PNAAssay):
        """Initialize a TooManyN pipeline predicate.

        :param count: the cutoff for the N count.
            If it is below 1.0, it will be considered a proportion, and above and equal to
            1 will be considered as discarding reads with a number of N's greater than this cutoff.
        :param assay: the assay configuration.
        """
        assert count >= 0
        self.is_proportion = count < 1.0
        self.cutoff = count
        self.assay = assay
        uei_pos = get_position_in_parent(self.assay, "uei")
        self._region1 = slice(0, uei_pos[0])
        self._region2 = slice(uei_pos[1], -1)

    def __repr__(self):
        """Return a string representation of the object."""
        return f"TooManyN(cutoff={self.cutoff}, is_proportion={self.is_proportion})"

    def test(self, read, info: ModificationInfo):
        """Return True if the read has too many N bases.

        :param read: the read to test
        :param info: the modification info
        """
        r = read.sequence.lower()
        region1 = r[self._region1]
        region2 = r[self._region2]

        read_len = len(region1) + len(region2)
        n_count = region1.count("n") + region2.count("n")
        if self.is_proportion:
            if len(read) == 0:
                return False
            return n_count / read_len > self.cutoff
        else:
            return n_count > self.cutoff



"""Module for PixelDataset and associated functions.

Copyright © 2023 Pixelgen Technologies AB.
"""

from __future__ import annotations

import logging
import warnings
from typing import (
    Any,
    Dict,
    Literal,
    Optional,
)

import numpy as np
import pandas as pd
import polars as pl
from anndata import AnnData

from pixelator.common.types import PathType
from pixelator.mpx.graph import Graph
from pixelator.mpx.pixeldataset.backends import (
    FileBasedPixelDatasetBackend,
    ObjectBasedPixelDatasetBackend,
    PixelDatasetBackend,
)
from pixelator.mpx.pixeldataset.datastores import (
    PixelDataStore,
    ZipBasedPixelFileWithCSV,
    ZipBasedPixelFileWithParquet,
)
from pixelator.mpx.pixeldataset.precomputed_layouts import PreComputedLayouts
from pixelator.mpx.pixeldataset.utils import (
    _enforce_edgelist_types,
    update_metrics_anndata,
)

logger = logging.getLogger(__name__)

SIZE_DEFINITION = "molecules"
# minimum number of vertices (nodes) required for polarization/co-localization
MIN_VERTICES_REQUIRED = 100


def read(path: PathType) -> PixelDataset:
    """Read a PixelDataset from a provided .pxl file.

    :param path: path to the file to read
    :return: an instance of `PixelDataset`
    :rtype: PixelDataset
    """
    return PixelDataset.from_file(path)


class PixelDataset:
    """PixelDataset represents data from one or more Molecular Pixelation experiments.

    In general the way to instantiate a new PixelDataset is to use it's `from_file`
    method:

    ```
    pxl_data = PixelDataset.from_file("/path/to/file.pxl")
    ```

    adata: the AnnData object containing marker counts from the various
           components -> required
    edgelist: the edge list (pd.DataFrame). This can be used to extract spatial
              graph information for each component -> required
    metadata: a dictionary including meta information -> Optional
    polarization: the polarization scores (pd.DataFrame) for
                  each component -> Optional
    colocalization: the colocalization scores (pd.DataFrame) for each
                    component -> Optional
    precomputed_layouts: pre-computed layouts that can be used to visualize
                         that component -> Optional
    """

    def __init__(self, backend: PixelDatasetBackend) -> None:
        """Create a `PixelDataset` from a `PixelDatasetBackend`.

        In general you should probably use `PixelDataset.from_file` to
        create a new `PixelDataset` instance.

        This method is reserved for advanced use.
        :param backend: an instance of `PixelDatasetBackend`
        """
        self._backend = backend

    @staticmethod
    def from_file(path: PathType) -> PixelDataset:
        """Create a new instance of `PixelDataset` from the file at the provided path.

        :param path: path to a .pxl file
        :return: A new instance of `PixelDataset`
        :rtype: PixelDataset
        """
        # We can ignore the error here, since while MyPy thinks that the
        # @cached_property decorators that we use on FileBasedPixelDatasetBackend
        # do not support setting, they actually do. See the documentation here:
        # https://docs.python.org/3/library/functools.html#functools.cached_property
        return PixelDataset(backend=FileBasedPixelDatasetBackend(path))  # type: ignore

    @staticmethod
    def from_data(
        adata: AnnData,
        edgelist: Optional[pd.DataFrame],
        metadata: Optional[Dict[str, Any]] = None,
        polarization: Optional[pd.DataFrame] = None,
        colocalization: Optional[pd.DataFrame] = None,
        precomputed_layouts: PreComputedLayouts | None = None,
        copy: bool = True,
        allow_empty_edgelist: bool = False,
    ) -> PixelDataset:
        """Create a new instance of PixelDataset from the provided underlying objects.

        :param adata: an instance of `AnnData`
        :param edgelist: an edgelist as a `pd.DataFrame`
        :param metadata: an instance of a dictionary with metadata, defaults to None
        :param polarization: a `pd.DataFrame` with polarization information,
                             defaults to None
        :param colocalization: a `pd.DataFrame` with colocalization information,
                               defaults to None
        :param precomputed_layouts: a PreComputedLayouts object, defaults to None
        :param copy: specify if the input data should be copied or not.
                     Defaults to True.
        :param allow_empty_edgelist: allow the edgelist to be empty. Defaults to False.
        :return: An instance of PixelDataset
        :rtype: PixelDataset
        """
        return PixelDataset(
            backend=ObjectBasedPixelDatasetBackend(
                adata=adata,
                edgelist=edgelist,
                metadata=metadata,
                polarization=polarization,
                colocalization=colocalization,
                precomputed_layouts=precomputed_layouts,
                copy=copy,
                allow_edgelist_to_be_empty=allow_empty_edgelist,
            )
        )

    @property
    def adata(self) -> AnnData:
        """Get the AnnData object."""
        return self._backend.adata

    @adata.setter
    def adata(self, value: AnnData) -> None:
        """Set the AnnData object."""
        self._backend.adata = value

    @property
    def edgelist(self) -> pd.DataFrame:
        """Get the edge list."""
        return _enforce_edgelist_types(self._backend.edgelist)

    @edgelist.setter
    def edgelist(self, value: pd.DataFrame) -> None:
        """Set the edge list."""
        self._backend.edgelist = _enforce_edgelist_types(value)

    @property
    def edgelist_lazy(self) -> pl.LazyFrame:
        """Get the edge list as a lazy dataframe."""
        lz_edgelist = self._backend.edgelist_lazy
        if "index" in set(lz_edgelist.collect_schema().names()):
            warnings.warn(
                "A column called `index` was identified in your edgelist. "
                "This will be removed."
            )
            lz_edgelist = lz_edgelist.drop("index")
        return lz_edgelist

    @property
    def polarization(self) -> Optional[pd.DataFrame]:
        """Get the polarization object."""
        return self._backend.polarization

    @polarization.setter
    def polarization(self, value: pd.DataFrame) -> None:
        """Set the polarization object."""
        self._backend.polarization = value

    @property
    def colocalization(self) -> Optional[pd.DataFrame]:
        """Get the colocalization object."""
        return self._backend.colocalization

    @colocalization.setter
    def colocalization(self, value: pd.DataFrame) -> None:
        """Set the colocalization object."""
        self._backend.colocalization = value

    @property
    def metadata(self) -> Dict:
        """Get the metadata dictionary."""
        metadata = self._backend.metadata
        if not metadata:
            return {}
        return metadata

    @metadata.setter
    def metadata(self, value: Dict) -> None:
        """Set the metadata object."""
        self._backend.metadata = value

    @property
    def precomputed_layouts(
        self,
    ) -> PreComputedLayouts:
        """Get the precomputed layouts."""
        return self._backend.precomputed_layouts

    @precomputed_layouts.setter
    def precomputed_layouts(self, value: PreComputedLayouts | None) -> None:
        """Set the precomputed layouts."""
        # Note that the type ignore here is to handle the fact that the setter
        # needs to be able to take None (in order to make it easier to the user)
        # but that will be transformed into a empty PreComputedLayouts object
        # by the backend as and when it sees fit.
        self._backend.precomputed_layouts = value  # type: ignore

    def graph(
        self,
        component_id: Optional[str] = None,
        add_node_marker_counts: bool = True,
        simplify: bool = True,
        use_full_bipartite: bool = True,
    ) -> Graph:
        """Get the graph from the underlying edgelist.

        :param component_id: Optionally give the component id of the component
                             to only return that component.
        :param add_node_marker_counts: Add marker counts to the nodes of the graph
        :param simplify: If True, removes self-loops and multiple edges between nodes
                         from the graph
        :param use_full_bipartite: If True, the full bipartite graph will be used,
                                   otherwise it will return the A-node projection
        :return: A Graph instance
        :rtype: Graph
        :raises: KeyError if the provided `component_id` is not found in the edgelist
        """
        if component_id:
            potential_component = self.edgelist_lazy.filter(
                pl.col("component").cast(pl.String) == component_id
            )
            if potential_component.head(1).collect().is_empty():
                raise KeyError(f"{component_id} not found in edgelist")
            return Graph.from_edgelist(
                potential_component,
                add_marker_counts=add_node_marker_counts,
                simplify=simplify,
                use_full_bipartite=use_full_bipartite,
            )
        return Graph.from_edgelist(
            self.edgelist_lazy,
            add_marker_counts=add_node_marker_counts,
            simplify=simplify,
            use_full_bipartite=use_full_bipartite,
        )

    def __str__(self) -> str:
        """Get a string representation of this object."""
        nbr_of_edges = self.edgelist_lazy.select(pl.count()).collect()[0, 0]
        msg = (
            f"Pixel dataset contains:\n"
            f"\tAnnData with {self.adata.n_obs} obs and {self.adata.n_vars} vars\n"
            f"\tEdge list with {nbr_of_edges} edges"
        )

        if self.polarization is not None:
            msg += f"\n\tPolarization scores with {self.polarization.shape[0]} elements"

        if self.colocalization is not None:
            msg += (
                "\n\tColocalization scores with "
                f"{self.colocalization.shape[0]} elements"
            )

        if (
            self.precomputed_layouts is not None
            and not self.precomputed_layouts.is_empty
        ):
            msg += "\n\tContains precomputed layouts"

        if self.metadata is not None:
            msg += "\n\tMetadata:\n"
            msg += "\n".join(
                [f"\t\t{key}: {value}" for key, value in self.metadata.items()]
            )

        msg = msg.replace("\t", "  ")

        return msg

    def __repr__(self) -> str:
        """Get a string representation of this object."""
        return str(self)

    def copy(self) -> PixelDataset:
        """Create a copy of the current PixelDataset instance.

        :return: A copy of the PixelDataset instance
        :rtype: PixelDataset
        """
        return PixelDataset.from_data(
            adata=self.adata.copy(),
            edgelist=self.edgelist.copy(),
            polarization=(
                self.polarization.copy() if self.polarization is not None else None
            ),
            colocalization=(
                self.colocalization.copy() if self.colocalization is not None else None
            ),
            metadata=self.metadata.copy() if self.metadata is not None else None,
            precomputed_layouts=(
                self.precomputed_layouts.copy()
                if self.precomputed_layouts is not None
                and not self.precomputed_layouts.is_empty
                else None
            ),
        )

    def save(
        self,
        path: PathType,
        file_format: Literal["csv", "parquet"] | PixelDataStore = "parquet",
        force_overwrite: bool = False,
    ) -> None:
        """Save the PixelDataset to a .pxl file in the location provided in `path`.

        :param path: the path where to save the dataset as a .pxl
        :param file_format: should be 'csv' or 'parquet'. Default is 'parquet'.
                            This indicates what file-format is used to serialize
                            the data frames in the .pxl file.
        :param force_overwrite: By default pixelator will not overwrite existing .pxl files, set this to true to
                                force an overwrite of the existing file.
        :returns: None
        :rtype: None
        :raises: AssertionError if invalid file format specified
        """
        logger.debug("Saving PixelDataset to %s", path)

        if isinstance(file_format, PixelDataStore):
            format_spec = file_format
        elif file_format not in ["csv", "parquet"]:
            raise AssertionError("`file_format` must be `csv` or `parquet`")
        if file_format == "csv":
            format_spec = ZipBasedPixelFileWithCSV(path)
        if file_format == "parquet":
            format_spec = ZipBasedPixelFileWithParquet(path)

        format_spec.save(self, force_overwrite=force_overwrite)

    def filter(
        self,
        components: Optional[pd.Series] = None,
        markers: Optional[pd.Series] = None,
    ) -> PixelDataset:
        """Filter the PixelDataset by components, markers, or both.

        Please note that markers will be filtered from the edgelist, even when provided,
        since it might cause different components to form, and hence invalidated the
        rest of the data.

        Consequently precomputed layouts will also not be filtered by marker.

        :param components: The components you want to keep, defaults to None
        :param markers: The markers you want to keep, defaults to None
        :return: A new instance of PixelDataset with the components/markers selected
        :rtype: PixelDataset
        """
        change_components = components is not None
        change_markers = markers is not None

        def _all_true_array(shape):
            return np.full(shape, True)

        adata_component_mask = (
            self.adata.obs.index.isin(components)
            if change_components
            else _all_true_array(self.adata.obs.index.shape)
        )
        adata_marker_mask = (
            self.adata.var.index.isin(markers)
            if change_markers
            else _all_true_array(self.adata.var.index.shape)
        )
        adata = self.adata[adata_component_mask, adata_marker_mask]
        update_metrics_anndata(adata, inplace=True)

        with pl.StringCache():
            lz_df = self.edgelist_lazy
            edgelist_pred = (
                lz_df.filter(pl.col("component").is_in(set(components)))  # type: ignore
                if change_components
                else self.edgelist_lazy
            )
            edgelist = _enforce_edgelist_types(edgelist_pred.collect().to_pandas())

        if self.polarization is not None:
            polarization_mask = (
                self.polarization["component"].isin(components)
                if change_components
                else _all_true_array(self.polarization.index.shape)
            ) & (
                self.polarization["marker"].isin(markers)
                if change_markers
                else _all_true_array(self.polarization.index.shape)
            )
            polarization = self.polarization[polarization_mask]

        if self.colocalization is not None:
            colocalization_mask = (
                self.colocalization["component"].isin(components)
                if change_components
                else _all_true_array(self.colocalization.index.shape)
            ) & (
                self.colocalization["marker_1"].isin(markers)
                & self.colocalization["marker_2"].isin(markers)
                if change_markers
                else _all_true_array(self.colocalization.index.shape)
            )
            colocalization = self.colocalization[colocalization_mask]

        if not (self.precomputed_layouts is None or self.precomputed_layouts.is_empty):
            precomputed_layouts = self.precomputed_layouts.filter(
                component_ids=components
            )

        return PixelDataset.from_data(
            adata=adata,
            edgelist=edgelist,
            metadata=self.metadata,
            polarization=polarization if self.polarization is not None else None,
            colocalization=colocalization if self.colocalization is not None else None,
            precomputed_layouts=(
                precomputed_layouts
                if not (
                    self.precomputed_layouts is None
                    or self.precomputed_layouts.is_empty
                )
                else None
            ),
        )



"""Helper classes to define the data input for the qc report.

Copyright © 2022 Pixelgen Technologies AB.
"""

import dataclasses
from typing import List, Optional, TypedDict, TypeVar

from pixelator.pna.report.models.commands import CommandInfo

CommandOptionValue = TypeVar("CommandOptionValue", str, int, float, bool)


@dataclasses.dataclass
class SampleInfo:
    """Dataclass for general sample information.

    The values defined here are used in the dynamic qc report to populate the
    "sample information" panel.
    """

    #: Version of pixelator used to generate the metrics
    pixelator_version: str

    #: Date the report was generated
    generation_date: str

    #: Identifier of the sample
    sample_id: str

    #: Short description of the sample
    sample_description: str

    #: Name of the pixelator panel file
    panel_name: Optional[str]

    #: Version of the pixelator panel file
    panel_version: Optional[str]

    #: Name of the technology used
    technology: Optional[str]

    #: Version of the technology used
    technology_version: Optional[str]

    #: List of commands and their parameters
    parameters: List[CommandInfo]


# --- START AUTOGENERATED: metrics ---
class Metrics(TypedDict, total=False):
    """A typed dictionary containing all metrics that can be used in the qc report.

    Values defined here will be used in the dynamic qc report to populate
    the different data tables. Missing values will be left blank.

    Note: This class is autogenerated from the metrics.json file (0.1.0)
          in the pixelator-qc-report-pna package.
    """

    #: Total number of read pairs that were input to this library.
    number_of_reads: int

    #: Fraction of total input reads discarded by the pipeline over all steps.
    fraction_discarded_reads: float

    #: Number of read pairs that were input to the amplicon assembly step.
    input_reads_amplicon: int

    #: Number of read pairs that were output from the amplicon assembly step.
    output_reads_amplicon: int

    #: Number of reads discarded during the amplicon assembly step.
    discarded_reads_amplicon: int

    #: Percentage of the total input reads discarded during the amplicon step. Reads
    #: can be discarded by amplicon when the (partial) LBS-1 and LBS-2 regions cannot
    #: be recognised.
    fraction_discarded_reads_amplicon: float

    #: Number of read pairs that were input to the demux step.
    input_reads_demux: int

    #: Number of reads that were error-corrected during the demux step.
    corrected_reads_demux: float

    #: Number of read pairs that were output from the demux step.
    output_reads_demux: int

    #: Percentage of reads discarded during the demux step. Reads can be discarded by
    #: demux when the PID1 and PID2 regions do not match the markers from the panel
    #: and cannot be corrected unambiguously.
    discarded_reads_demux: float

    #: Fraction of total input reads discarded by demux because of invalid marker
    #: barcodes.
    fraction_discarded_reads_demux: float

    #: Number of read pairs that were the input to the collapse step.
    input_reads_collapse: int

    #: Number of reads that were the output from the collapse step.
    output_reads_collapse: int

    #: Number of unique molecules that were the output from the collapse step.
    output_molecules_collapse: int

    #: The total number of Reads that were recovered by error correction in the
    #: collapse step.
    corrected_reads_collapse: float

    #: Fraction of total input reads with invalid PID1 & PID2 after error-correction.
    fraction_of_molecules_corrected_collapse: float

    #: Number of read pairs that were the input to the graph step.
    input_reads_graph: int

    #: Number of read pairs that were included in the final graph.
    output_reads_graph: int

    #: Number of molecules that were the input to the graph step.
    input_molecules_graph: int

    #: Number of molecules that were included in the final graph.
    output_molecules_graph: int

    #: Number of molecules discarded during the graph step.
    discarded_reads_graph: int

    #: Number of molecules discarded during the graph step.
    discarded_molecules_graph: int

    #: Fraction of total input reads discarded during the graph step.
    fraction_discarded_reads_graph: float

    #: The redundancy of valid reads. Valid reads are reads that have passed the
    #: demux step. The saturation is calculated as: <br /> $$ 1- rac{	ext{\#
    #: unique valid reads}}{	ext{\# total valid reads}} $$
    valid_reads_saturation: float

    #: The redundancy of edge detection. The saturation is calculated as: <br /> $$ 1
    #: - rac{	ext{\# graph edges}}{	ext{\# graph reads}} $$
    graph_edge_saturation: float

    #: The redundancy of node detection. The saturation is calculated as: <br /> $$ 1
    #: - rac{	ext{\# graph nodes}}{ 	ext{\# graph reads}} $$
    graph_node_saturation: float

    #: Percentage of total input reads that are valid.
    fraction_valid_reads: float

    #: Percentage of total input reads in valid components.
    fraction_graph_reads: float

    #: Fraction of bases in a full amplicon with Q-score ≥ 30, excluding very low
    #: quality/no-call (Q ≤ 2) bases from the denominator.
    fraction_q30_bases: float

    #: Fraction of bases in UMI1 with Q-score ≥ 30, excluding very low quality/no-
    #: call (Q ≤ 2) bases from the denominator.
    fraction_q30_bases_in_umi1: float

    #: Fraction of bases in UMI2 with Q-score ≥ 30, excluding very low quality/no-
    #: call (Q ≤ 2) bases from the denominator.
    fraction_q30_bases_in_umi2: float

    #: Fraction of bases in PID1 with Q-score ≥ 30, excluding very low quality/no-
    #: call (Q ≤ 2) bases from the denominator.
    fraction_q30_bases_in_pid1: float

    #: Fraction of bases in PID2 with Q-score ≥ 30, excluding very low quality/no-
    #: call (Q ≤ 2) bases from the denominator.
    fraction_q30_bases_in_pid2: float

    #: Fraction of bases in UEI with Q-score ≥ 30, excluding very low quality/no-call
    #: (Q ≤ 2) bases from the denominator.
    fraction_q30_bases_in_uei: float

    #: Fraction of bases in LBS1 with Q-score ≥ 30, excluding very low quality/no-
    #: call (Q ≤ 2) bases from the denominator.
    fraction_q30_bases_in_lbs1: float

    #: Fraction of bases in LBS2 with Q-score ≥ 30, excluding very low quality/no-
    #: call (Q ≤ 2) bases from the denominator.
    fraction_q30_bases_in_lbs2: float

    #: The number of cells annotated by the pipeline.
    number_of_cells: int

    #: Median of the number of usable reads per cell.
    median_reads_per_cell: float

    #: Median of the total number of markers per cell.
    median_markers_per_cell: float

    #: Connectivity expressed as the median of the average k-coreness of the cells.
    median_average_k_coreness: float

    #: Fraction of the total input reads that are in cells.
    fraction_discarded_cells: float

    #: The fraction of cells that are outliers as determined by the tau-metric
    #: outlier detection.
    fraction_of_outlier_cells: float


# --- END AUTOGENERATED: metrics ---


class InfoAndMetrics(TypedDict):
    """Aggregated info and metrics."""

    info: SampleInfo
    metrics: Metrics


@dataclasses.dataclass
class QCReportData:
    """Collection of data inputs for metrics and dynamic figures and tables in the qc report.

    All attributes are strings containing csv or json data ready to be
    injected into the dynamic report.
    """

    metrics: Metrics

    #: CSV formatted string containing a DataFrame for the
    #: "component_size_distribution" figure
    #: component_size, rank, selected, markers
    #: 18463, 1, True, 32
    #: 18290, 2, True, 39
    ranked_component_size: Optional[str]

    #: CSV formatted string containing a DataFrame used for plotting embeddings
    #: with various colors,
    #: eg (cell type, leiden cluster, umis)
    #: Following columns are required:
    #:  component,umap1,umap2,umis,cluster,cluster_cell_class
    #: For example:
    #: ```
    #: component,umap1,umap2,cluster,cluster_cell_class,reads,molecules,mean_upia_degree,mean_umi_per_upia
    #: 0,11.891996,12.216717,4793,2,other
    #: 1,11.02211,12.486254,6708,2,other
    #: 2,7.8869944,7.5836325,5031,5,predict_B-cells
    #: 3,0.66538113,4.00765,3446,9,predict_CD4+_T-cells
    #: 4,13.21712,5.2930307,6704,1,predict_B-cells
    #: ```
    component_data: Optional[str]

    #: CSV formatted string containing a DataFrame used to generate the antibodies
    #: per cell figure.
    #: Following columns are required: antibody,percentage
    #: For example:
    #: ```
    #: antibody,percentage
    #: CD274,0.0016104640666519112
    #: CD44,0.04755881148351282
    #: CD56,0.020829214318265954
    #: ```
    antibody_percentages: Optional[str]

    #: CSV formatted string containing a DataFrame used to generate the antibody
    #: counts violin plots.
    #: It consists of the component column followed by a list of antibody columns.
    #: For example:
    #: ```
    #: component,CD274,CD44,CD56
    #: PXLCMP0000000,9.0,76.0,63.0
    #: PXLCMP0000002,3.0,90.0,166.0
    #: PXLCMP0000003,1.0,44.0,113.0
    #: ```
    antibody_counts: Optional[str]

    #: CSV formatted string containing a DataFrame used to generate proximity heatmaps
    #:
    #: Following columns are required:
    #:  marker_1,marker_2,join_count_z,log2ratio
    #:
    #: For example:
    #: ```
    #: marker_1,marker_2,join_count_z,log2ratio
    #: CD45,HLA-ABC,0.02165450650176937,0.008440362800652843
    #: HLA-ABC,HLA-ABC,0.20508868569666797,0.010748605487355567
    #: HLA-ABC,TCRab,-0.289434564966545,-0.28095631383105635
    #: ACTB,CD45,0.0,0.0
    #: ACTB,HLA-ABC,0.23253978095569725,0.21996568394190796
    #: CD3e,CD3e,0.7555555555555555,0.6844981742720706
    proximity_data: Optional[str]

    def to_dict(self) -> dict[str, Optional[str]]:
        """Convert the dataclass to a dictionary."""
        return dataclasses.asdict(self)



"""Helper functions to collect data for report generation.

Copyright © 2022 Pixelgen Technologies AB.
"""

from __future__ import annotations

import itertools
import json
import logging
import typing

import numpy as np
import numpy.typing as npt
import pandas as pd
import polars as pl
import scanpy as sc

from pixelator.common.utils.simplification import simplify_line_rdp
from pixelator.pna.pixeldataset import PNAPixelDataset
from pixelator.pna.report.common import WorkdirOutputNotFound
from pixelator.pna.report.qcreport.types import Metrics, QCReportData

logger = logging.getLogger(__name__)


if typing.TYPE_CHECKING:
    from anndata import AnnData

    from pixelator.pna.graph.report import GraphSampleReport
    from pixelator.pna.report.common import PixelatorPNAReporting


def collect_components_umap_data(adata: AnnData) -> str:
    """Create a csv formatted string with the components umap data for the qc report.

    :param adata: an AnnData object with the umap data (obs)
    :return: a csv formatted string with umap data
    :rtype: str
    """
    empty = np.full(adata.n_obs, np.nan)

    sc.pp.neighbors(
        adata,
        use_rep="clr",
        n_neighbors=min(adata.n_obs, 15),
        random_state=1,
    )

    init_pos: typing.Literal["random", "spectral"] = (
        "random" if adata.n_obs - 1 <= 2 else "spectral"
    )
    sc.tl.umap(adata, min_dist=0.5, n_components=2, random_state=1, init_pos=init_pos)

    if "X_umap" in adata.obsm:
        umap_df = pd.DataFrame(
            adata.obsm["X_umap"], columns=["umap1", "umap2"], index=adata.obs.index
        )
    else:
        umap_df = pd.DataFrame({"umap1": empty, "umap2": empty}, index=adata.obs.index)

    umap_df["cluster"] = empty

    umap_df = pd.concat(
        [
            umap_df,
            adata.obs[["reads_in_component", "n_antibodies", "n_umi", "n_edges"]],
        ],
        axis=1,
    )
    return umap_df.to_csv(index=True, index_label="component")


def collect_antibody_percentages_data(adata: AnnData) -> str:
    """Create the antibody percentages histogram data for the qc report.

    This function created a csv formatted string with the antibody name and the
    percentage of the antibody aggregated over all components.

    :param adata: an AnnData object with antibody counts percentages data
    :return: a csv formatted string with antibody percentages data
    :rtype: str
    """
    index = adata.var.index.set_names("antibody", inplace=False)
    df = pd.DataFrame(
        {"count": adata.var["antibody_count"], "percentage": adata.var["antibody_pct"]},
        index=index,
    )
    return df.to_csv()


def collect_antibody_counts_data(adata: AnnData) -> str:
    """Create the antibody counts data for the qc report.

    :param adata: an AnnData object with the antibody counts data
    :return: a csv formatted string with the antibody counts data
    :rtype: str
    """
    return adata.to_df().to_csv(index=True, index_label="component")


def collect_ranked_component_size_data(
    adata: AnnData, graph_report: GraphSampleReport, subsample: bool = True
) -> str:
    """Create data for the `cell calling` and `component size distribution` plot.

    This collects the component size and the number of antibodies per component.
    Components that pass the filters (is_filtered) are marked as selected.

    :param adata: The AnnData object with the component metrics
    :param graph_report: The graph report object
    :param subsample_non_cell_comoponents: if True, subsample non-cell components
    :param subsample_epsilon: the epsilon value for the subsampling.
        Ignored if subsample_non_cell_comoponents is False
    :return: a csv formatted string with the plotting data
    :rtype: str
    """
    data = adata.obs

    sorted_keys = sorted(
        list(graph_report.pre_filtering_component_sizes.keys()), reverse=True
    )
    component_sizes = np.fromiter(
        itertools.chain.from_iterable(
            itertools.repeat(k, graph_report.pre_filtering_component_sizes[k])
            for k in sorted_keys
        ),
        dtype=int,
    )
    df = pd.DataFrame({"size": component_sizes})
    df["selected"] = np.logical_and(
        df["size"] >= (graph_report.component_size_min_filtering_threshold or 0),
        df["size"] <= (graph_report.component_size_max_filtering_threshold or np.inf),
    )

    df.sort_values(by="size", ascending=False, inplace=True)
    df["rank"] = df["size"].rank(method="average", ascending=False)

    if subsample:
        cell_mask = df["selected"].to_numpy()
        coords = np.ascontiguousarray(df[["rank", "size"]].to_numpy())
        cell_idx = np.flatnonzero(cell_mask)
        other_coords_mask = ~cell_mask
        other_coords_idx = np.flatnonzero(other_coords_mask)
        other_coords = coords[other_coords_mask]

        if len(other_coords) != 0:
            simplified_coords_idx = simplify_line_rdp(
                other_coords, 1e-2, return_mask=True
            )

            # Check for non empty here since zero length simplified_coords_idx
            # has shape conflicts when concatenating
            global_coords = np.concatenate(
                [cell_idx, other_coords_idx[simplified_coords_idx]]
            )
            global_coords.sort()

            df = df.iloc[global_coords]
            # These should still be sorted since we used sorted indices
            # but lets sort them again by rank just to make sure
            df.sort_values(by="rank", inplace=True)

    df.set_index("rank", inplace=True)
    return df.to_csv(index=True)


def _build_symmetric_proximity_matrix(
    df: pl.DataFrame, labels: npt.NDArray
) -> npt.NDArray:
    """Build a symmetric matrix from the pivoted dataframe.

    The dataframe is expected to have a column 'marker_1' and the rest of the columns
    are the markers.

    The matrix is built by filling the missing values with 0, sorting the columns and
    rows according to a fixed order, and then forcing the matrix to be symmetric
    by averaging it with its transpose.

    Params:
        df: The dataframe to build the matrix from
        labels: The labels to use for the fixed order of the matrix

    Returns:
        The symmetric matrix as a numpy array

    """
    missing_cols = set(labels) - set(df.columns[1:])
    missing_rows = set(labels) - set(df["marker_1"])

    # Create a mapping dictionary for the fixed order
    order_mapping = {value: index for index, value in enumerate(labels)}

    # Add missing columns
    for col in missing_cols:
        df = df.with_columns(pl.lit(None).alias(col))

    # Add missing rows
    for row in missing_rows:
        df = df.vstack(
            pl.DataFrame({"marker_1": [row], **{col: None for col in df.columns[1:]}})
        )

    assert df.shape[0] == df.shape[1] - 1, (
        "Unexpected shape after correcting for missing rows/columns"
    )

    # Sort the columns starting from index 1
    sorted_columns = sorted(
        df.columns[1:], key=lambda col: order_mapping.get(col, float("inf"))
    )

    # Reorder columns
    df_sorted = df.select(["marker_1"] + sorted_columns)
    # Reorder rows
    df_sorted = df_sorted.sort(by=pl.col("marker_1").replace_strict(order_mapping))

    assert (df_sorted["marker_1"] == df_sorted.columns[1:]).all(), (
        "Row and column order do not match"
    )

    matrix = df_sorted.drop("marker_1").to_numpy()

    # Replace nan values by the matching transposed values to make a symmetric matrix
    # and fill the rest with 0
    symmetric_matrix = np.where(np.isnan(matrix), matrix.T, matrix)
    symmetric_matrix = np.nan_to_num(symmetric_matrix, copy=False, nan=0.0)

    assert not np.isnan(symmetric_matrix).any(), (
        "Matrix still contains NaN values after filling"
    )

    return symmetric_matrix


def collect_proximity_data(
    pxl_dataset: PNAPixelDataset,
    min_marker_count: int = 50,
    min_join_count_expected_mean=2,
    min_component_count=2,
    max_markers=50,
) -> str | None:
    """Collect the proximity data for the qc report.

    Params:
        pxl_dataset: The PNAPixelDataset object
        min_marker_count: The minimum marker count for a marker pair to be included in the median scores
        min_join_count_expected_mean: The minimum join count expected mean for a marker pair to be included in the median scores
        min_component_count:
            The minimum number of components needed for a marker pair after previous filtering
            to be included in the median scores.
        max_markers:
            The maximum number of markers to include. Markers with the highest stddev are picked.d
    Returns:
        A json formatted string with the proximity data or None of no proximity data is found.
    """
    df = pxl_dataset.proximity().to_polars()
    if df is None:
        return None

    # Filter the scores
    filtered_df = df.filter(
        (pl.col("marker_1_count") >= min_marker_count)
        & (pl.col("marker_2_count") >= min_marker_count)
        & (pl.col("join_count_expected_mean") >= min_join_count_expected_mean)
    )

    # Group by marker_1, marker_2 and count the number of components that
    # have that pair after previously filtering for marker count
    group_counts = filtered_df.group_by(["marker_1", "marker_2"]).len()

    # Filter the dataframe to only include the pairs that have at least min_component_count items
    filtered_df = (
        filtered_df.join(group_counts, on=["marker_1", "marker_2"], how="left")
        .filter(pl.col("len") > min_component_count)
        .drop("len")
    )

    # Group by marker_1, marker_2 and take the median across components for join_count_z and log2_ratio
    result_df = filtered_df.group_by(["marker_1", "marker_2"]).agg(
        [
            pl.median("join_count_z").alias("median_z_score"),
            pl.median("log2_ratio").alias("median_log2ratio"),
        ]
    )

    # Create unique sorted array of markers
    labels = np.unique(group_counts[["marker_1", "marker_2"]].to_numpy())

    join_count_z = result_df.pivot(
        on="marker_2",
        index="marker_1",
        values="median_z_score",
        aggregate_function="median",
        sort_columns=True,
    )

    n_labels = len(labels)
    join_count_z_matrix = _build_symmetric_proximity_matrix(join_count_z, labels)

    log2_ratio = result_df.pivot(
        on="marker_2",
        index="marker_1",
        values="median_log2ratio",
        aggregate_function="median",
        sort_columns=True,
    )

    log2_ratio_matrix = _build_symmetric_proximity_matrix(log2_ratio, labels)

    join_count_z_std = np.std(join_count_z_matrix, axis=1)
    log2_ratio_std = np.std(log2_ratio_matrix, axis=1)

    # This might not make much sense, but it is just to define a total order
    combined_std = join_count_z_std + log2_ratio_std

    # Select the last N element
    marker_subset = np.sort(np.argsort(combined_std)[n_labels - max_markers : n_labels])

    filtered_labels = labels[marker_subset]
    filtered_idx = np.ix_(marker_subset, marker_subset)
    join_count_z_matrix_filtered = join_count_z_matrix[filtered_idx]
    log2_ratio_matrix_filtered = log2_ratio_matrix[filtered_idx]

    return json.dumps(
        {
            "markers": filtered_labels.tolist(),
            "join_count_z": np.round(join_count_z_matrix_filtered, decimals=4).tolist(),
            "log2_ratio": np.round(log2_ratio_matrix_filtered, decimals=4).tolist(),
        }
    )


def collect_metrics_report_data(
    reporting: PixelatorPNAReporting, sample_id: str
) -> Metrics:
    """Collect the data needed for the main metrics in the qc report.

    The `annotate` folder must be present in `input_path`.

    :param reporting: The report object to use for data collection from the workdir
    :param sample_id: The sample id
    :returns Metrics: A Metrics object
    """
    logger.debug(
        "Collecting QC report data for %s in %s", sample_id, reporting.workdir.basedir
    )

    amplicon_metrics = reporting.amplicon_metrics(sample_id)
    demux_metrics = reporting.demux_metrics(sample_id)
    collapse_metrics = reporting.collapse_metrics(sample_id)
    graph_metrics = reporting.graph_metrics(sample_id)
    analysis_metrics = reporting.analysis_metrics(sample_id)
    reads_flow = reporting.reads_flow(sample_id)

    def _fraction_or_zero(x: int | float, y: int | float) -> float:
        return x / y if y > 0.0 else 0.0

    # Global level metrics
    # TODO: Move them into reporting.reads_flow ?

    total_input_reads = amplicon_metrics.input_reads
    fraction_discarded_reads = _fraction_or_zero(
        (amplicon_metrics.input_reads - graph_metrics.reads_output), total_input_reads
    )
    fraction_discarded_reads_amplicon = _fraction_or_zero(
        amplicon_metrics.total_failed_reads, total_input_reads
    )
    fraction_discarded_reads_demux = _fraction_or_zero(
        demux_metrics.failed_reads, total_input_reads
    )
    fraction_discarded_reads_graph = _fraction_or_zero(
        graph_metrics.discarded_reads, total_input_reads
    )
    valid_reads_saturation = 1 - _fraction_or_zero(
        collapse_metrics.output_molecules, collapse_metrics.input_reads
    )
    fraction_valid_reads = _fraction_or_zero(
        collapse_metrics.input_reads, total_input_reads
    )
    fraction_graph_reads = _fraction_or_zero(
        graph_metrics.reads_output, total_input_reads
    )

    # Map pixelator metrics to QC report metrics
    metrics = Metrics(
        # Read level statistics
        number_of_reads=reads_flow.input_read_count,
        # Amplicon
        input_reads_amplicon=amplicon_metrics.input_reads,
        output_reads_amplicon=amplicon_metrics.output_reads,
        discarded_reads_amplicon=amplicon_metrics.total_failed_reads,
        # Demux
        input_reads_demux=demux_metrics.input_reads,
        output_reads_demux=demux_metrics.output_reads,
        corrected_reads_demux=collapse_metrics.corrected_reads,
        discarded_reads_demux=demux_metrics.failed_reads,
        # Collapse
        input_reads_collapse=collapse_metrics.input_reads,
        output_reads_collapse=collapse_metrics.output_reads,
        output_molecules_collapse=collapse_metrics.output_molecules,
        corrected_reads_collapse=collapse_metrics.corrected_reads,
        # Graph
        input_molecules_graph=graph_metrics.molecules_input,
        output_molecules_graph=graph_metrics.molecules_output,
        input_reads_graph=graph_metrics.reads_input,
        output_reads_graph=graph_metrics.reads_output,
        discarded_reads_graph=graph_metrics.discarded_reads,
        discarded_molecules_graph=graph_metrics.discarded_molecules,
        # Discarded reads percentages vs total reads
        fraction_discarded_reads=fraction_discarded_reads,
        fraction_discarded_reads_amplicon=fraction_discarded_reads_amplicon,
        fraction_discarded_reads_demux=fraction_discarded_reads_demux,
        fraction_discarded_reads_graph=fraction_discarded_reads_graph,
        # Saturation
        graph_node_saturation=graph_metrics.node_saturation,
        graph_edge_saturation=graph_metrics.edge_saturation,
        valid_reads_saturation=valid_reads_saturation,
        # Global "read conversion" into useful data
        fraction_valid_reads=fraction_valid_reads,
        fraction_graph_reads=fraction_graph_reads,
        # Global and per region Q30 statistics
        fraction_q30_bases_in_pid1=amplicon_metrics.q30_statistics.pid1,
        fraction_q30_bases_in_pid2=amplicon_metrics.q30_statistics.pid2,
        fraction_q30_bases_in_umi1=amplicon_metrics.q30_statistics.umi1,
        fraction_q30_bases_in_umi2=amplicon_metrics.q30_statistics.umi2,
        fraction_q30_bases_in_lbs1=amplicon_metrics.q30_statistics.lbs1,
        fraction_q30_bases_in_lbs2=amplicon_metrics.q30_statistics.lbs2,
        fraction_q30_bases_in_uei=amplicon_metrics.q30_statistics.uei,
        fraction_q30_bases=amplicon_metrics.q30_statistics.total,
        number_of_cells=graph_metrics.component_count_post_component_size_filtering,
        median_reads_per_cell=graph_metrics.median_reads_per_component,
        median_markers_per_cell=graph_metrics.median_markers_per_component,
        median_average_k_coreness=(
            analysis_metrics.k_cores.median_average_k_core
            if analysis_metrics.k_cores
            else 0.0
        ),
        fraction_of_outlier_cells=graph_metrics.fraction_of_aggregate_components,
    )

    logger.debug(
        "QC report data collected for %s in %s", sample_id, reporting.workdir.basedir
    )
    return metrics


def collect_report_data(
    reporting: PixelatorPNAReporting, sample_id: str
) -> QCReportData:
    """Collect the data needed to generate figures in the qc report.

    :param reporting: The PixelatorPNAReporting object
    :param sample_id: The sample id to collect data for
    :returns QCReportData: A QCReportData object
    """
    logger.debug(
        "Collecting QC report data for %s in %s", sample_id, reporting.workdir.basedir
    )

    workdir = reporting.workdir
    # parse filtered dataset
    dataset_file = workdir.filtered_dataset(sample_id)
    dataset = PNAPixelDataset.from_files(dataset_file)
    adata = dataset.adata()
    graph_report = reporting.graph_metrics(sample_id)
    metric_data = collect_metrics_report_data(reporting, sample_id)

    component_data = collect_components_umap_data(adata)
    antibody_percentages = collect_antibody_percentages_data(adata)
    antibody_counts = collect_antibody_counts_data(adata)
    ranked_component_sizes = collect_ranked_component_size_data(adata, graph_report)

    try:
        analysis_dataset_file = workdir.analysed_dataset(sample_id)
        analysis_dataset = PNAPixelDataset.from_files(analysis_dataset_file)
        proximity_data = collect_proximity_data(analysis_dataset)
    except WorkdirOutputNotFound:
        logger.warning(
            "Analysis stage dataset not found for %s. Skipping proximity data in report.",
            sample_id,
        )
        proximity_data = None

    # build the report data
    data = QCReportData(
        metrics=metric_data,
        component_data=component_data,
        ranked_component_size=ranked_component_sizes,
        antibody_percentages=antibody_percentages,
        antibody_counts=antibody_counts,
        proximity_data=proximity_data,
    )

    logger.debug("QC report data collected for %s in %s", sample_id, workdir.basedir)
    return data



"""Helpers for configuration plugin entrypoints.

Copyright © 2023 Pixelgen Technologies AB.
"""

from __future__ import annotations

import importlib.metadata
import logging
from importlib.metadata import EntryPoint
from typing import TYPE_CHECKING, Generator, List, Union

try:
    from importlib.metadata import EntryPoints
except ImportError:
    # Python 3.8 and 3.9
    pass

if TYPE_CHECKING:
    from pixelator.common.config import Config

logger = logging.getLogger(__name__)


def fetch_config_plugins() -> Generator[EntryPoint, None, None]:
    """Find plugins and return them as in a generator.

    :yields EntryPoint: The entrypoint object
    :returns: A generator of the loaded plugins
    """
    eps = importlib.metadata.entry_points()
    group = "pixelator.config_plugin"
    selected_entrypoints: Union[List[EntryPoint], EntryPoints]

    if hasattr(eps, "select"):
        # New interface in Python 3.10 and newer versions of the
        # importlib_metadata backport.
        selected_entrypoints = eps.select(group=group)
    else:
        # Older interface, deprecated in Python 3.10 and recent
        # importlib_metadata, but we need it in Python 3.8 and 3.9.
        selected_entrypoints = eps.get(group, [])  # type: ignore

    for entrypoint in selected_entrypoints:
        logger.debug("Detected config plugin %s", entrypoint.name)
        yield entrypoint


def load_config_plugins(config: Config) -> Config:
    """Load all config plugins."""
    new_config = config

    for entry_point in fetch_config_plugins():
        logger.debug("Loading config plugin %s", entry_point.name)
        new_config = entry_point.load()(new_config)

    return new_config



"""Console script for pixelator layout creation.

Copyright © 2024 Pixelgen Technologies AB.
"""

import logging
import sys
from pathlib import Path

import click

from pixelator.common.utils import (
    get_sample_name,
    log_step_start,
    sanity_check_inputs,
    timer,
    write_parameters_file,
)
from pixelator.pna.analysis_engine import (
    AnalysisManager,
    LoggingSetup,
)
from pixelator.pna.cli.common import output_option
from pixelator.pna.layout import CreateLayout
from pixelator.pna.pixeldataset import PxlFile, read
from pixelator.pna.report.common import PixelatorPNAWorkdir
from pixelator.pna.report.models.layout import LayoutSampleReport

logger = logging.getLogger(__name__)


@click.command(
    "layout",
    short_help="compute graph layouts that can be used to visualize components",
    options_metavar="<options>",
)
@click.argument(
    "pxl_file",
    nargs=1,
    required=True,
    type=click.Path(exists=True),
    metavar="PXL",
)
@click.option(
    "--layout-algorithm",
    required=False,
    multiple=True,
    default=["wpmds_3d"],
    help="Select a layout algorithm to use. This can be specified multiple times to compute multiple layouts. Default: pmds_3d",
    type=click.Choice(["pmds_3d", "wpmds_3d"]),
)
@click.option(
    "--pmds-pivots",
    default=50,
    required=False,
    type=click.IntRange(50, None),
    show_default=True,
    help="Number of pivots to use for the PMDS layout algorithm. Default: 50. More give better results, but increase computation times.",
)
@click.option(
    "--wpmds-k",
    default=3,
    required=False,
    type=click.IntRange(1, 10),
    show_default=True,
    help="The window size used when computing probability weights to the wpmds layout method. Only used when the wpmds layout method is selected. Default: 3.",
)
@output_option
@click.pass_context
@timer
def layout(
    ctx,
    pxl_file,
    layout_algorithm,
    pmds_pivots,
    wpmds_k,
    output,
):
    """Compute graph layouts that can be used to visualize components."""
    log_step_start(
        "layout",
        input_files=pxl_file,
        layout_algorithm=layout_algorithm,
        pmds_pivots=pmds_pivots,
        wpmds_k=wpmds_k,
        output=output,
    )

    # some basic sanity check on the input files
    sanity_check_inputs(pxl_file, allowed_extensions="pxl")

    # create output folder if it does not exist
    workdir = PixelatorPNAWorkdir(output)
    layout_output_dir = workdir.stage_dir("layout")

    logger.info(f"Computing layout(s) for file {pxl_file}")

    clean_name = get_sample_name(pxl_file)
    write_parameters_file(
        ctx,
        layout_output_dir / f"{clean_name}.meta.json",
        command_path="pixelator single-cell-pna layout",
    )

    analysis_to_run = [
        CreateLayout(
            layout_algorithm,
        )
    ]

    pxl_file = PxlFile(Path(pxl_file))
    pxl_dataset = read(pxl_file.path)

    logging_setup = LoggingSetup.from_logger(ctx.obj.get("LOGGER"))
    analysis_manager = AnalysisManager(analysis_to_run, logging_setup=logging_setup)
    pxl_file_target = PxlFile.copy_pxl_file(
        pxl_file, layout_output_dir / f"{clean_name}.layout.pxl"
    )
    pxl_dataset = analysis_manager.execute_from_path(
        input_pxl_file_path=pxl_file.path, pxl_file_target=pxl_file_target
    )

    metrics_file = layout_output_dir / f"{clean_name}.report.json"
    report = LayoutSampleReport(
        sample_id=clean_name,
        product_id="single-cell-pna",
        report_type="layout",
    )
    report.write_json_file(metrics_file, indent=4)


if __name__ == "__main__":
    sys.exit(layout())



"""Module with helpers for calculating the adjacency matrix of a graph.

Given a list of close neigbours from a search on a FAISS index.

Copyright © 2024 Pixelgen Technologies AB.
"""

import numpy as np
import numpy.typing as npt
import scipy


def build_network_directional(
    distances: npt.NDArray[np.int64],
    indices: npt.NDArray[np.uint64],
    read_counts: npt.NDArray[np.int64],
    cutoff: int,
) -> scipy.sparse.csr_matrix:
    """Create a sparse directed adjacency matrix from a FAISS k-nearest query result.

    Uses the "Directional" approach as described by UMI-tools.

    :param distances: The distances of the k-nearest neighbors.
        A 2D array with shape (n_queries, k).
    :param indices: The indices of the k-nearest neighbors.
        A 2D array with shape (n_queries, k).
    :param read_counts: The read support for each molecule.
        A 1D array with shape (n_queries,).
    :param cutoff: The distance cutoff to consider an edge between two nodes.
    :return: A sparse directed adjacency matrix in CSR format.
    :rtype: scipy.sparse.csr_matrix
    """
    if distances.shape != indices.shape:
        raise ValueError(
            f"The distances and indices matrix must have the same shape."
            f"D({distances.shape}) != I({indices.shape})"
        )

    n_queries = distances.shape[0]

    # Hamming distance must be less than or equal to the cutoff
    # Each nucleotide difference will result in a hamming distance of 2
    # The cutoff is already in hamming space so the caller must have adjusted for this
    edges = distances <= cutoff

    # Retrieve the non-zero indices for the candidate edges
    i, j = np.nonzero(edges)  # type: ignore

    # row indices so we can reuse those directly as the node_a index
    node_a_index = i
    # retrieve the column (node_b) indices from the Indices matrix from the FAISS query result
    node_b_index = indices[i, j]

    # Retrieve read counts for the edge candidates
    node_a_read_counts = read_counts[node_a_index]
    node_b_read_counts = read_counts[node_b_index]

    # Only keep edges where the read counts are at least twice as high for node A
    nodes_mask = node_a_read_counts >= (2 * node_b_read_counts - 1)

    row = node_a_index[nodes_mask]
    col = node_b_index[nodes_mask]
    data = np.ones(len(row), dtype=np.uint8)

    # Build a CSR matrix from the adjacency information in COO format
    directional_adjacency = scipy.sparse.coo_matrix(
        (data, (row, col)), shape=(n_queries, n_queries), dtype=np.uint8
    )
    # Convert to CSR format for efficient connected components calculation later
    directional_adjacency = directional_adjacency.tocsr()
    return directional_adjacency


# Note: read_counts is not used in this function, but we keep it here to
# make the signature the same as build_network_directional


def build_network_cluster(
    distances: npt.NDArray[np.int64],
    indices: npt.NDArray[np.uint64],
    read_counts: npt.NDArray[np.int64],
    cutoff: int,
) -> scipy.sparse.csr_matrix:
    """Create a sparse adjacency matrix from a FAISS k-nearest query result.

    Use the "Cluster" approach as described by UMI-tools.

    :param distances: The distances of the k-nearest neighbors.
        A 2D array with shape (n_queries, k).
    :param indices: The indices of the k-nearest neighbors.
        A 2D array with shape (n_queries, k).
    :param read_counts: The read support for each molecule.
        A 1D array with shape (n_queries,).
    :param cutoff: The distance cutoff to consider an edge between two nodes.
    """
    n_queries = distances.shape[0]

    offsets = np.zeros((n_queries + 1), dtype=np.uint64)
    edges = distances <= cutoff
    edge_weights = np.ones_like(distances[edges])
    edge_indices = indices[edges]

    # Build the offsets array to directly construct a CSR matrix
    np.cumsum(np.count_nonzero(edges, axis=1), out=offsets[1:])

    adjacency = scipy.sparse.csr_matrix(
        (edge_weights, edge_indices, offsets), shape=(n_queries, n_queries)
    )

    return adjacency


def build_network_cluster_from_range_query(lims, distances, indices, index_map=None):
    """Create a sparse adjacency matrix from a FAISS range query result.

    FAISS range queries already returns a CSR matrix

    So for query i:
     - lims[i], lims[i+1] is the range in I and D that lists the neighbors of i
     - I[lims[i]:lims[i+1]] are the indices from the database vectors
     - D[lims[i]:lims[i+1]] are the corresponding distances


    :param lims: The range limits for the neighbors of each query.
    :param distances: The distances of the neighbors.
    :param indices: The indices of the neighbors.
    :param index_map: A transformation of the indices
    """
    indices = indices

    if index_map:
        indices = index_map[indices]

    # The FAISS result is actually identical to the CSR sparse matrix format
    # We can just load it into a scipy sparse matrix without copying where the data is stored

    adjacency = scipy.sparse.csr_matrix(
        (distances, indices, lims), shape=(len(indices), len(indices))
    )
    return adjacency



"""Workflow test helper for single-cell report command.

Copyright © 2023 Pixelgen Technologies AB.
"""

import logging
from pathlib import Path

import pytest

from .base import BaseWorkflowTestMixin

logger = logging.getLogger(__name__)


class BaseReportTestsMixin(BaseWorkflowTestMixin):
    """Base class for report command tests.

    Test cases (defined in this class or in subclasses)
    that depend on the output should be marked with:
    ```
    @pytest.mark.dependency(depends=["test_report_run"])
    ```
    """

    __stage_key__ = "report"

    @pytest.mark.dependency(name="test_report_run", depends=["test_analysis_run"])
    def test_report_run(self):
        """Test and run the report command."""
        params = self.__get_parameters()
        verbose = self.__get_options("common").get("verbose")
        panel = self.__get_data("panel")
        panel_file = self.__get_data("panel_file")

        command = [
            "--log-file",
            "report-pixelator.log",
            "single-cell-mpx",
            "report",
            ".",
            "--output",
            ".",
            "--panel",
            str(Path(panel_file).name) if panel_file else panel,
        ]

        if verbose:
            command = ["--verbose"] + command

        if params:
            command += params

        file_deps = []
        if panel_file is not None:
            file_deps.append(panel_file)

        self.context.run_command("report", command, file_deps)

        assert self.__this_result.exit_code == 0

    @pytest.mark.dependency(depends=["test_report_run"])
    def test_report_qc_report_exists(self):
        """Check if the qc report html exists."""
        report_files = list((self.workdir / "report").glob("*.qc-report.html"))
        assert len(report_files) > 0



"""Types associated with colocalization.

Copyright © 2023 Pixelgen Technologies AB.
"""

from dataclasses import dataclass
from typing import Callable, Literal

import pandas as pd

from pixelator.mpx.analysis.types import RegionByCountsDataFrame

MarkerColocalizationResults = pd.DataFrame

TransformationTypes = Literal["raw", "log1p", "rate-diff"]


@dataclass
class CoLocalizationFunction:
    """Container for colocalization functions and their names."""

    name: str
    func: Callable[  # type: ignore
        [RegionByCountsDataFrame], MarkerColocalizationResults  # type: ignore
    ]  # type: ignore



"""Combine collapse reports and parquet files into a single report and parquet file.

Copyright © 2025 Pixelgen Technologies AB
"""

import json
import logging
from pathlib import Path, PurePath
from typing import Iterable

import pyarrow.parquet as pq

from pixelator.pna.collapse.paired.statistics import (
    CollapseStatistics,
    MarkerLinkGroupStats,
)

logger = logging.getLogger(__name__)


def combine_parquet_files(input_files: Iterable[Path], output_file: Path) -> Path:
    """Scan a directory for parquet files and stream them into a single parquet file.

    Args:
        input_files: The folder containing the parquet files.
        output_file: The path to the output parquet file.

    Returns:
        The path to the combined parquet file. This is the same path as `output_file`.

    """
    files = list(input_files)
    first_file = pq.read_table(files.pop(0))

    with pq.ParquetWriter(output_file, first_file.schema) as writer:
        writer.write_table(first_file)

        for f in files:
            table = pq.read_table(f)
            writer.write_table(table)

    return output_file


def combine_report_files(input_files: Iterable[Path]) -> CollapseStatistics:
    """Combine a list of JSON collapse report files into a single report.

    :param input_files: The folder containing the parquet files.
    :returns: A new CollapseStatistics instance containing the combined report data.
    """
    files = list(input_files)
    combined_stats = CollapseStatistics()

    for report_file in files:
        with open(report_file, "r") as f:
            json_data = json.load(f)

            per_marker_stats = json_data.pop("markers")
            collapse_data = json_data

            # TODO: Remove this path once the unnecessary nesting is removed from the JSON
            if (
                isinstance(per_marker_stats, list)
                and len(per_marker_stats) == 1
                and isinstance(per_marker_stats[0], list)
            ):
                per_marker_stats = per_marker_stats[0]

            # Add markers
            for marker_group in per_marker_stats:
                m = MarkerLinkGroupStats.model_validate(marker_group)
                combined_stats.add_marker_stats(
                    m.marker_1,
                    m.marker_2,
                    cluster_stats=m,
                )

            for file_data in collapse_data["processed_files"]:
                combined_stats.add_input_file(
                    PurePath(file_data["path"]),
                    molecule_count=file_data["molecule_count"],
                    file_size=file_data["file_size"],
                )

    return combined_stats



"""Copyright © 2023 Pixelgen Technologies AB."""

from pixelator.pna.config.config_class import (
    PNAConfig,
    load_assays_package,
    load_panels_package,
)
from pixelator.pna.config.plugin import load_config_plugins

pna_config = PNAConfig()

assays_package = "pixelator.pna.resources.assays"
pna_config = load_assays_package(pna_config, assays_package)

panels_package = "pixelator.pna.resources.panels"
pna_config = load_panels_package(pna_config, panels_package)

pna_config = load_config_plugins(pna_config)



"""Copyright © 2023 Pixelgen Technologies AB."""

from typing import Any, ClassVar, List

from click.testing import Result as CliRunnerResult


class BaseWorkflowTestMixin:
    """Base class for WorkflowTestMixin classes.

    These mixin classes are always used in conjunction with the PixlatorWorkflowTest.
    """

    __stage_key__: ClassVar[str]

    # Type hints for these methods that are injected through __init_subclass__.
    def __get_data(self, key: str) -> Any:
        """Try to retrieve a key from various data sources.

        Following sources are tried in order:

        1. Stage specific data
        2. Common data
        3. Test config data
        """
        ...

    def __get_parameters(self) -> List[str]:  # type: ignore
        """Retrieve the parameters for this stage."""
        ...

    def __get_options(self, key: str) -> Any:  # type: ignore
        """Retrieve a key from the options objects."""
        ...

    def __get_common_data(self, key: str) -> Any:  # type: ignore
        """Retrieve a key from the common data."""
        ...

    @property
    def __this_result(self) -> CliRunnerResult:  # type: ignore
        """Retrieve the execution result object for this stage."""
        ...

    @property
    def __this_command(self) -> List[str]:  # type: ignore
        """Retrieve the CLI object for this stage."""
        ...

    @property
    def __this_logs(self) -> str:  # type: ignore
        """Retrieve the logs of this stage if set."""
        ...

    def __init_subclass__(cls, **kwargs):
        """Some hackery to bind data access functions."""
        super().__init_subclass__(**kwargs)
        thisclass = BaseWorkflowTestMixin
        clsname = cls.__name__
        datakey = cls.__stage_key__

        def get_parameters_helper(instance):
            return thisclass.__base_get_parameters(instance, datakey)

        def get_data_helper(instance, key):
            return thisclass.__base_get_data(instance, key, data_key=datakey)

        def get_options_helper(instance, key):
            return thisclass.__base_get_options(instance, key)

        def get_common_options_helper(instance, key):
            return thisclass.__base_get_common_options(instance, key)

        setattr(cls, f"_{clsname}__get_data", get_data_helper)

        setattr(cls, f"_{clsname}__get_parameters", get_parameters_helper)

        setattr(cls, f"_{clsname}__get_options", get_options_helper)

        setattr(cls, f"_{clsname}__get_common_options", get_common_options_helper)

        def get_this_result(self):
            return self.context.results[datakey]

        def get_this_command(self):
            return self.context.commands[datakey]

        def get_this_logs(self):
            return self.context.logs[datakey]

        setattr(cls, f"_{clsname}__this_result", property(get_this_result))
        setattr(cls, f"_{clsname}__this_command", property(get_this_command))
        setattr(cls, f"_{clsname}__this_logs", property(get_this_logs))

    def __base_get_data(self, name: str, *, data_key=None) -> Any:
        """Retrieve undefined attributes from the class variable dicts."""
        return getattr(self, name, None)

    def __base_get_options(self, key: str):
        """Retrieve test config data from the class variables defined in the test class."""
        if not hasattr(self, "options"):
            raise ValueError("No options section found in test config.")

        options = getattr(self, "options")
        return options.get(key)

    def __base_get_common_options(self, key: str):
        if not hasattr(self, "options"):
            raise ValueError("No options section found in test config.")

        options = getattr(self, "options")
        common = options.get("common")
        return common or []

    def __base_get_parameters(self, stage):
        if not hasattr(self, "options"):
            raise ValueError("No options section found in test config.")

        options = getattr(self, "options")
        stage_data = options.get(stage)
        stage_params = stage_data.get("params") if stage_data else None

        return stage_params or []



"""Module containing utilities for simulating data.

Module containing utilities for simulating data. in testing, and in-silico experiments.
Copyright © 2023 Pixelgen Technologies AB.
"""

from typing import Generator, Iterable, Iterator, List, Optional

import numpy as np

from pixelator.common.config import Assay
from pixelator.mpx.config.config_instance import config


class ReadSimulator:
    """Simulate reads for tests and experiments.

    Simplistic read simulator that produces reads based on a assay configuration.
    It works by generating template molecule, that is then "sequenced" by sampling
    it a number of times. It allows for the addition of base substitutions errors
    with a given probability to simulate simple sequencing errors.
    """

    _ERRORS_DICT = {
        "A": ["C", "G", "T"],
        "C": ["A", "G", "T"],
        "G": ["C", "A", "T"],
        "T": ["C", "G", "A"],
    }

    def __init__(
        self,
        assay: Optional[Assay] = None,
        upia_pool_size: int = 1000,
        upib_pool_size: int = 1000,
        random_seed: int = 42,
        markers: List[str] = ["AAAAAAAA"],
    ) -> None:
        """Create a ReadSimulator instance.

        :param assay: an `Assay` to use will use D21 if None is passed,
                      defaults to None
        :param nbr_of_a_pixels: The number of A pixels to sample molecules from,
                                defaults to 1000
        :param nbr_of_b_pixels: The number of B pixels to sample molecules from,
                                defaults to 1000
        :param random_seed: Set the random seed, defaults to 42
        :param markers: A list of marker barcodes to use in generating the reads,
                        defaults to ["AAAAAAAA"]
        """
        if not assay:
            assay = config.assays["D21"]
        self.assay = assay
        self.rng = np.random.default_rng(random_seed)
        self.upia_pool = [self.random_seq(25) for _ in range(upia_pool_size)]
        self.upib_pool = [self.random_seq(25) for _ in range(upib_pool_size)]
        self.markers = markers

    def random_seq(self, length: int) -> str:
        """Generate a random DNA sequence of size length.

        :param length: length of the sequence to generate
        :return: A random DNA sequence
        :rtype: str
        """
        return "".join([self.rng.choice(["A", "T", "C", "G"]) for _ in range(length)])

    def build_molecule(self, nbr_of_molecules: int) -> Generator[str, None, None]:
        """Create strings representing molecules, that can then be "sequenced".

        :param nbr_of_molecules: number of molecules to generate
        :yields str: An iterator of DNA sequences
        :rtype: Generator[str, None, None]
        """

        def build_fragment():
            for region in self.assay.get_region_by_id("amplicon").regions:
                if region.region_id == "upi-b":
                    yield self.rng.choice(self.upib_pool)
                if region.region_id == "pbs-2":
                    yield region.sequence
                if region.region_id == "upi-a":
                    yield self.rng.choice(self.upia_pool)
                if region.region_id == "pbs-1":
                    yield region.sequence
                if region.region_id == "umi-b":
                    yield self.random_seq(region.max_len)
                if region.region_id == "bc":
                    yield self.rng.choice(self.markers)

        for _ in range(nbr_of_molecules):
            yield "".join(build_fragment())

    def sequence_molecule(
        self,
        molecules: Iterable[str],
        mean_nbr_of_reads_per_molecule: float,
        std_nbr_of_reads_per_molecule: float,
    ) -> Generator[str, None, None]:
        """Simulate sequencing of reads from a set of molecules.

        Generate "reads" of the input molecules, the number of times
        each molecule is read depends on `mean_nbr_of_reads_per_molecule`
        and `std_nbr_of_reads_per_molecule`.

        :param molecules: the underlying molecules to "sequence"
        :param mean_nbr_of_reads_per_molecule: the mean number of reads to generate per
                                               molecule
        :param std_nbr_of_reads_per_molecule: the standard deviation in the number of
                                              reads generated per molecule
        :yields str: an iterator of "sequenced" reads
        :rtype: Generator[str, None, None]
        """
        for molecule in molecules:
            nbr_of_times_sequences = int(
                self.rng.normal(
                    mean_nbr_of_reads_per_molecule, std_nbr_of_reads_per_molecule
                )
            )
            for _ in range(nbr_of_times_sequences):
                yield molecule

    def add_sequencing_errors(
        self, reads: Iterable[str], error_prob_per_base: float
    ) -> Generator[str, None, None]:
        """Add sequencing errors to reads.

        Add base substitutions errors to `reads` with a probability
        given by `error_prop_per_base`.

        :param reads: reads to add errors to
        :param error_prob_per_base: probability of adding an error, range: [0,1)
        :yields str: an iterator of DNA sequences with errors added to it
        :raises: Assertion error if `error_prob_per_base` is invalid
        :rtype: Generator[str, None, None]
        """
        if error_prob_per_base < 0 or error_prob_per_base > 1:
            raise AssertionError("`error_prob_per_base` must be between 0 and 1.")

        def add_errors(read):
            def data():
                for base in read:
                    if self.rng.random() < error_prob_per_base:
                        yield self.rng.choice(self._ERRORS_DICT[base])
                    else:
                        yield base

            return "".join(data())

        for read in reads:
            yield add_errors(read)

    def simulated_reads(
        self,
        n_molecules: int,
        mean_reads_per_molecule: float,
        std_reads_per_molecule: float,
        prob_of_seq_error: Optional[float] = 0,
    ) -> Iterator[str]:
        """Simulate a reads from a pool n molecules, adding sequencing errors.

        :param n_molecules: number of molecules to start from
        :param mean_reads_per_molecule: mean number of molecules to generate
                                        per molecule
        :param std_reads_per_molecule: the standard deviation of the number of
                                       molecules to generate
        :param prob_of_seq_error: probability of base substitutions errors per base
                                  sequenced. Default: 0.
        :return: An iterator of sequence reads
        :rtype: Iterator[str]
        """
        fragments = self.build_molecule(n_molecules)
        reads = self.sequence_molecule(
            fragments,
            mean_nbr_of_reads_per_molecule=mean_reads_per_molecule,
            std_nbr_of_reads_per_molecule=std_reads_per_molecule,
        )
        if prob_of_seq_error:
            return self.add_sequencing_errors(
                reads, error_prob_per_base=prob_of_seq_error
            )
        return reads



"""Common functions for collapsing data.

Copyright © 2024 Pixelgen Technologies AB.
"""

import dataclasses
import typing
from pathlib import Path
from typing import Generator, Iterable

import faiss
import numpy as np
import scipy
from numpy import typing as npt


def build_binary_index(data) -> faiss.IndexBinary:
    """Build a binary index for a batch of embeddings.

    The input data is a matrix of shape (n, d) where n is the number
    of embeddings and d is the dimensionality of each embedding.

    Depending on the size of the input data, either a flat or HNSW index is built.

    :param data: The data to build the index for.
    """
    """Build a binary index for the given data."""
    db_len = data.shape[0]
    d = data.shape[1] * 8

    if db_len < 5_000:
        index = faiss.IndexBinaryFlat(d)
    else:
        index = faiss.IndexBinaryHNSW(d, 16)

    # Add data all at once to utilise fast internal parallelization in FAISS
    index.add(data)

    return index


class FAISSBackend:
    """A backend for building and searching binary indexes using FAISS."""

    def __init__(self, threads: int | None = None):
        """Initialize the FAISS Backend.

        Args:
            threads: The number of threads to use. -1 or None will use all available threads

        """
        if threads is not None and threads >= 1:
            faiss.omp_set_num_threads(threads)

    def build_index(self, data: np.ndarray) -> faiss.IndexBinary:
        """Build a binary index for the given data.

        Args:
            data: The data to build the index for.

        Returns:
            The built index.

        """
        return build_binary_index(data)

    def search(self, index: faiss.IndexBinary, data: np.ndarray, k: int) -> np.ndarray:
        """Search the index for the k nearest neighbors of the given data.

        Args:
            index: The index to search.
            data: The data to search for.
            k: The number of nearest neighbors to find.

        Returns:
            The indices of the k nearest neighbors for each data point.
            Indices will be -1 when less than k neighbors are found.

        """
        return index.search(data, k)


def _collect_label_array_indices(
    labels: npt.NDArray[np.int32], n_components: int
) -> npt.NDArray[np.object_]:
    """Collect an array with a label for each item into an array of indices for each label.

    :param labels: the labels for each item
    :param n_components: Number of labels in the label array
    """
    # Single pass over the labels to collect the indices of each connected component
    groups = np.ndarray(shape=(n_components,), dtype=object)  # type: ignore

    for idx, value in enumerate(labels):
        g = groups[value]
        if g is None:
            groups[value] = [idx]
        else:
            g.append(idx)

    return groups


def _split_chunks(
    n_components: int, chunk_size: int
) -> Generator[tuple[int, int], None, None]:
    """Split a range [0, n_components) into chunks of size chunk_size.

    This is a generator function that yields the start and stop indices of each chunk.
    The last chunk may be smaller than chunk_size.

    :param n_components: The size of the input range
    :param chunk_size: The size of each chunk
    :yield: A tuple of (start, stop) indices for each chunk
    """
    pos = 0
    complete_chunks = n_components // chunk_size

    for i in range(complete_chunks):
        yield (pos, pos + chunk_size)
        pos += chunk_size

    yield (pos, n_components)


class SplitFilesResult(typing.NamedTuple):
    """Tuple with a list per category of collapse input parquet files."""

    m1: list[Path]
    m2: list[Path]
    paired: list[Path]


def _split_files_per_marker_files(inputs: Iterable[Path | str]) -> SplitFilesResult:
    """Split the input files into separate lists based on the marker used.

    Args:
        inputs: The input files to split.

    Returns:
        A tuple of lists containing the input files for each marker.

    """
    umi1_files: list[Path] = []
    umi2_files: list[Path] = []
    paired_files: list[Path] = []

    for f in inputs:
        f = Path(f)
        if ".m1." in f.name:
            umi1_files.append(f)
        elif ".m2." in f.name:
            umi2_files.append(f)
        else:
            paired_files.append(f)

    return SplitFilesResult(umi1_files, umi2_files, paired_files)


def split_collapse_inputs(
    parquet_files: Iterable[Path | str],
) -> list[Path] | tuple[list[Path], list[Path]]:
    """Check the input parquet files to determine independent or paired collapsed data.

    Detection is simply based on the presence of ".m1." or ".m2." in the file name.

    :param parquet_files: The parquet files to check.
    :returns: Either a single list when the input data has been collapsed using the "paired" strategy,
        or a tuple of two lists when collapsed using "independent" strategy.
    :raises ValueError: If the input data contains a mix of independent and paired collapsed data.
    """
    # Check parquet files first
    umi1_files, umi2_files, paired_files = _split_files_per_marker_files(parquet_files)

    is_independent_collapsed = len(umi1_files) > 0 and len(umi2_files) > 0
    is_paired_collapsed = len(paired_files) > 0

    if is_paired_collapsed and is_independent_collapsed:
        raise ValueError(
            "Cannot combine from demux/collapse stage with different `--strategy` in the same run."
        )

    if is_paired_collapsed:
        return paired_files

    return umi1_files, umi2_files


MoleculeCollapserAlgorithm = typing.Literal["directional", "cluster"]


def _find_connected_components_directional(
    csgraph,
) -> tuple[int, npt.NDArray[np.int32]]:
    """Find connected components in a graph returned by the directional collapse algorithm.

    :param csgraph: The sparse adjacency matrix of the connected components
    """
    # Return the (weakly) connected components in a directed sparse graph
    n_components, labels = scipy.sparse.csgraph.connected_components(
        csgraph, directed=True, connection="weak", return_labels=True
    )
    return n_components, labels


def _find_connected_components_cluster(
    csgraph,
) -> tuple[int, npt.NDArray[np.int32]]:
    """Find connected components in a graph returned by the "cluster" collapse algorithm.

    :param csgraph: The sparse adjacency matrix of the connected components
    """
    # Return the (weakly) connected components in a directed sparse graph
    n_components, labels = scipy.sparse.csgraph.connected_components(
        csgraph, directed=False, return_labels=True
    )
    return n_components, labels


@dataclasses.dataclass(frozen=True, slots=True)
class CollapseInputs:
    """Output object for :func:`check_collapse_strategy_inputs`."""

    parquet: list[Path] | tuple[list[Path], list[Path]]
    reports: list[Path] | tuple[list[Path], list[Path]]


def check_collapse_strategy_inputs(
    parquet_files: Iterable[Path | str], report_files: Iterable[Path | str]
) -> CollapseInputs:
    """Check the input parquet files to determine independent or paired collapsed data.

    Detection is simply based on the presence of ".m1." or ".m2." in the file name.

    :param parquet_files: The parquet files to check.
    :returns: Either a single list when the input data has been collapsed using the "paired" strategy,
        or a tuple of two lists when collapsed using "independent" strategy.
    :raises ValueError: If the input data contains a mix of independent and paired collapsed data.
    """
    # Check parquet input files
    umi1_files, umi2_files, paired_files = _split_files_per_marker_files(parquet_files)

    # Check report input files
    umi1_report_files, umi2_report_files, paired_report_files = (
        _split_files_per_marker_files(report_files)
    )

    # Sanity checks
    if len(umi1_files) != len(umi1_report_files):
        raise ValueError(
            "Mismatched number of UMI1 parquet files and UMI1 report files."
        )

    if len(umi2_files) != len(umi2_report_files):
        raise ValueError(
            "Mismatched number of UMI2 parquet files and UMI2 report files."
        )

    if len(paired_files) != len(paired_report_files):
        raise ValueError("Mismatched number of parquet files and report files.")

    # Determine independent or paired strategy from the input files
    is_independent_collapsed = len(umi1_files) > 0 and len(umi2_files) > 0
    is_paired_collapsed = len(paired_files) > 0

    if is_paired_collapsed and is_independent_collapsed:
        raise ValueError(
            "Cannot combine from demux/collapse stage with different `--strategy` in the same run."
        )

    if is_paired_collapsed:
        return CollapseInputs(parquet=paired_files, reports=paired_report_files)

    return CollapseInputs(
        parquet=(umi1_files, umi2_files), reports=(umi1_report_files, umi2_report_files)
    )



"""Graph module.

Functions and classes relating to PNA graphs.

Copyright © 2024 Pixelgen Technologies AB.
"""

import logging
from collections import defaultdict
from timeit import default_timer as timer
from typing import Iterable, Optional

import networkx as nx
import pandas as pd
import polars as pl
import pyarrow as pa

from pixelator.common.graph import Graph as BaseGraph
from pixelator.common.graph.backends.implementations._networkx import (
    NetworkXGraphBackend,
)
from pixelator.common.graph.backends.protocol import (
    SupportedLayoutAlgorithm,
    VertexClustering,
)

logger = logging.getLogger(__name__)


class PNAGraph(BaseGraph):
    """Graph class for PNA data."""

    def __init__(self, backend: NetworkXGraphBackend):
        """Create a new graph instance."""
        self._backend = backend
        self._connected_components_needs_recompute = False
        self._connected_components: VertexClustering | None = None

    def from_record_batches(edgelist: Iterable[pa.RecordBatch], **kwargs) -> "PNAGraph":
        """Create a graph from record batches."""
        return PNAGraph(PNAGraphBackend.from_record_batches(edgelist, **kwargs))

    def from_edgelist(edgelist: pl.LazyFrame, **kwargs):  # type: ignore
        """Create a graph from an edgelist."""
        return PNAGraph(PNAGraphBackend.from_edgelist(edgelist, **kwargs))

    @property
    def node_marker_counts(self):
        """Return the marker counts for each node."""
        return self._backend.node_marker_counts()

    def layout_coordinates(  # type: ignore
        self,
        layout_algorithm: SupportedLayoutAlgorithm = "wpmds_3d",
        get_node_marker_matrix: bool = True,
        random_seed: Optional[int] = None,
        **kwargs,
    ) -> pd.DataFrame:
        """Generate coordinates and (optionally) node marker counts for plotting.

        Generate a dataframe with coordinates, and (optionally) node marker
        counts to use that can be used for plotting.

        The layout options are:
          - fruchterman_reingold
          - fruchterman_reingold_3d
          - kamada_kawai
          - kamada_kawai_3d
          - pmds
          - pmds_3d
          - wpmds_3d


        The `pmds` options are much faster than the force-directed algorithms fruchterman_reingold
        and kamada_kawai. The `wpmds_3d` option is a weighted version of the `pmds_3d` algorithm.

        :param layout_algorithm: the layout algorithm to use to generate the coordinates
        :param get_node_marker_matrix: Add a matrix of marker counts to each
                                       node if True.
        :param random_seed: used as the seed for graph layouts with a stochastic
                            element. Useful to get deterministic layouts across
                            method calls.
        :param **kwargs: will be passed to the underlying layout implementation
        :return: the coordinates and markers (if activated) as a dataframe
        :rtype: pd.DataFrame
        :raises: AssertionError if the provided `layout_algorithm` is not valid
        :raises: ValueError if the provided current graph instance is empty
        """
        return self._backend.layout_coordinates(
            layout_algorithm=layout_algorithm,
            get_node_marker_matrix=get_node_marker_matrix,
            random_seed=random_seed,
            **kwargs,
        )


class PNAGraphBackend(NetworkXGraphBackend):
    """Graph backend for PNA data that reuses most of the NetworkXGraphBackend."""

    @staticmethod
    def _build_graph_from_row_iterator(
        g: nx.Graph, row_iterator: Iterable[dict], **kwargs
    ):
        read_count_per_node: dict[str, int] = defaultdict(int)
        node_marker_type: dict[str, str] = defaultdict(str)
        node_type: dict[str, str] = defaultdict(str)

        def create_edges():
            for row in row_iterator:
                node1, node2 = row["umi1"], row["umi2"]
                read_count_per_node[node1] += row["read_count"]
                read_count_per_node[node2] += row["read_count"]
                node_marker_type[node1] = row["marker_1"]
                node_marker_type[node2] = row["marker_2"]
                node_type[node1] = "A"
                node_type[node2] = "B"
                yield node1, node2

        g.add_edges_from(create_edges())
        nx.set_node_attributes(g, read_count_per_node, "read_count")
        nx.set_node_attributes(g, node_marker_type, "marker")
        nx.set_node_attributes(g, node_type, "pixel_type")

        node_names = {node: node for node in g.nodes()}
        nx.set_node_attributes(g, node_names, "name")
        return g

    @staticmethod
    def _build_graph_from_lazy_frame(g: nx.Graph, edgelist: pl.LazyFrame, **kwargs):
        return PNAGraphBackend._build_graph_from_row_iterator(
            g, edgelist.collect().iter_rows(named=True), **kwargs
        )

    @staticmethod
    def _build_graph_from_pandas(g: nx.Graph, edgelist: pd.DataFrame, **kwargs):
        return PNAGraphBackend._build_graph_from_row_iterator(
            g, [edge for _, edge in edgelist.iterrows()], **kwargs
        )

    @staticmethod
    def from_edgelist(edgelist: pl.LazyFrame | pd.DataFrame, **kwargs):  # type: ignore
        """Create a graph from an edgelist."""
        g: nx.Graph = nx.empty_graph(0, nx.Graph)
        if isinstance(edgelist, pl.LazyFrame):
            g = PNAGraphBackend._build_graph_from_lazy_frame(g, edgelist, **kwargs)
        elif isinstance(edgelist, pd.DataFrame):
            g = PNAGraphBackend._build_graph_from_pandas(g, edgelist, **kwargs)

        return PNAGraphBackend(g)

    @staticmethod
    def from_record_batches(batches: Iterable[pa.RecordBatch], **kwargs):
        """Create a graph from an edgelist."""
        # TODO This is completely untested!
        g: nx.Graph = nx.empty_graph(0, nx.Graph)
        for batch in batches:
            edgelist = batch.to_pylist()
            g = PNAGraphBackend._build_graph_from_row_iterator(g, edgelist, **kwargs)

        return PNAGraphBackend(g)

    def node_marker_counts(self) -> pd.DataFrame:
        """Return the marker counts for each node."""
        markers_df = pd.DataFrame.from_dict(
            nx.get_node_attributes(self.raw, "marker"), orient="index"
        )
        markers_df = markers_df.reset_index(names="node")
        markers_df = markers_df.pivot_table(
            index="node", columns=0, aggfunc="size", fill_value=0
        )
        markers_df.columns.name = "marker"

        return markers_df

    def layout_coordinates(  # type: ignore
        self,
        layout_algorithm: SupportedLayoutAlgorithm = "wpmds_3d",
        get_node_marker_matrix: bool = True,
        random_seed: Optional[int] = None,
        **kwargs,
    ) -> pd.DataFrame:
        """Generate coordinates and (optionally) node marker counts for plotting.

        Generate a dataframe with coordinates, and (optionally) node marker
        counts to use that can be used for plotting.

        The layout options are:
          - pmds
          - pmds_3d
          - fruchterman_reingold
          - fruchterman_reingold_3d
          - kamada_kawai
          - kamada_kawai_3d
          - wpmds_3d

        For most cases the `pmds` options should be about 10-100x faster
        than the force directed layout methods, i.e. `fruchterman_reingold`
        and `kamada_kawai`. Among the force directed layout methods,
        `fruchterman_reingold` is generally faster than `kamada_kawai`. The
        `wpmds_3d` method uses edge weights to improve the layout, but is slightly
        slower than `pmds_3d`.

        :param layout_algorithm: the layout algorithm to use to generate the coordinates
        :param get_node_marker_matrix: Add a matrix of marker counts to each
                                       node if True.
        :param random_seed: used as the seed for graph layouts with a stochastic
                            element. Useful to get deterministic layouts across
                            method calls.
        :param **kwargs: will be passed to the underlying layout implementation
        :return: the coordinates and markers (if activated) as a dataframe
        :rtype: pd.DataFrame
        :raises: AssertionError if the provided `layout_algorithm` is not valid
        :raises: ValueError if the provided current graph instance is empty
        """
        start_time = timer()
        coordinates = self._layout_coordinates(
            layout_algorithm=layout_algorithm, random_seed=random_seed, **kwargs
        )

        if get_node_marker_matrix:
            node_marker_counts = self.node_marker_counts().astype(pd.UInt8Dtype())  # type: ignore
            df = pd.concat([coordinates, node_marker_counts], axis=1)
        else:
            df = coordinates
        pixel_name_and_type = pd.DataFrame.from_records(
            [(v.index, v["pixel_type"]) for v in self.vs.vertices()],
            columns=["index", "pixel_type"],
        )
        pixel_name_and_type = pixel_name_and_type.set_index("index", drop=False)
        df = pd.concat([pixel_name_and_type, df], axis=1)

        logger.debug(
            "Layout computation using %s took %.2f seconds",
            layout_algorithm,
            timer() - start_time,
        )
        return df



"""Copyright © 2025 Pixelgen Technologies AB."""

from pixelator.common.config.assay import (
    Assay,
    AssayModel,
    Region,
    RegionModel,
    RegionType,
    SequenceType,
    get_position_in_parent,
)
from pixelator.common.config.config_class import Config, load_assays_package
from pixelator.common.config.panel import (
    AntibodyPanel,
    AntibodyPanelMetadata,
    load_antibody_panel,
)
from pixelator.common.config.utils import load_yaml_file

__all__ = [
    "AssayModel",
    "RegionModel",
    "RegionType",
    "SequenceType",
    "Assay",
    "Region",
    "get_position_in_parent",
    "Config",
    "load_assays_package",
    "load_yaml_file",
    "AntibodyPanel",
    "AntibodyPanelMetadata",
    "load_antibody_panel",
]



"""Functions for finding aggregates.

Copyright © 2022 Pixelgen Technologies AB.
"""

import logging
from typing import Optional

import numpy as np
from anndata import AnnData
from scipy.stats import iqr

from pixelator.common.annotate.constants import (
    TAU_HARD_THRESHOLD,
    TAU_IQR_LOWER_THRESHOLD,
    TAU_IQR_UPPER_THRESHOLD,
)

logger = logging.getLogger(__name__)


def specificity_tau(matrix: np.ndarray) -> np.ndarray:
    """Tau specificity score computed as described in [1]_.

    Essentially it gives us a score between 0 and 1, where a component that
    expresses a single marker would have a tau score of 1, and one where all
    markers are equally expressed would have a tau score of 0.

    .. [1] Yanai I, Benjamin H, Shmoish M, Chalifa-Caspi V, Shklar M, Ophir R,
        Bar-Even A, Horn-Saban S, Safran M, Domany E, Lancet D, Shmueli O.
        Genome-wide midrange transcription profiles reveal expression level
        relationships in human tissue specification. Bioinformatics.
        2005 Mar 1;21(5):650-9.
        doi: 10.1093/bioinformatics/bti042. Epub 2004 Sep 23. PMID: 15388519.

    :param matrix: a numpy matrix of marker counts
    :return: a vector of the computed tau values
    """
    max_count = np.max(matrix, axis=1)
    _, nbr_markers = matrix.shape
    with np.errstate(divide="ignore", invalid="ignore"):
        x_prim = np.divide(matrix, max_count[:, None])
    tau_results = np.nansum(1 - x_prim, axis=1) / (nbr_markers - 1)

    return tau_results


def call_aggregates(adata: AnnData, inplace: bool = True) -> Optional[AnnData]:
    """Call aggregates on the adata instance.

    We defined aggregates as components where either:
     - A single or a handful of markers account for almost all of the count data.
       These can likely be attributed to single antibodies forming aggregates
     - Low tau scores, meaning a an even number of counts for multiple markers.
       These likely come from multiple antibodies forming aggregates.

    For downstream analysis both of these types should be removed for most types
    of analysis.

    We find aggregates by computing a tau specificity score (see `tau_specificity`
    for details).

    We mark components as "high" if they that have a tau score above
    `annotation.constants.TAU_HARD_THRESHOLD`, or have a tau score above
    `annotation.constants.TAU_IQR_UPPER_THRESHOLD` * inter-quartile range
    from the median.

    We mark components as "low" if they that have a tau score below
    `annotation.contants.TAU_IQR_LOWER_THRESHOLD` * inter-quartile range
    from the median.

    The following data is added to the AnnData:
        - `obs["tau"]` = The tau specificity score of the component
        - `obs["tau_type"]` = "normal"/"high"/"low" for components with
            the respective levels of tau scores.
        - `uns["tau_thresholds"]["tau_upper_hard_limit"]` the upper hard
           limit used to set `tau_type` as high
        - `uns["tau_thresholds"]["tau_upper_iqr_limit]` the upper limit
           based on IQR used to set `tau_type` as high
        - `uns["tau_thresholds"]["tau_lower_iqr_limit]` the lower limit
           based on IQR used to set `tau_type` as low

    :param adata: an AnnData object to call aggregates on
    :param inplace: If `True` performs the operation inplace
    :return: the updated AnnData object or None if inplace is True.
    """
    logging.debug("Calling aggregates based on tau specificity scores")

    if not inplace:
        adata = adata.copy()

    tau_values = specificity_tau(adata.X)
    hard_limit = tau_values >= TAU_HARD_THRESHOLD
    iqr_upper_threshold = (
        np.median(tau_values) + iqr(tau_values) * TAU_IQR_UPPER_THRESHOLD
    )
    iqr_limit_upper = tau_values >= iqr_upper_threshold
    iqr_lower_threshold = (
        np.median(tau_values) - iqr(tau_values) * TAU_IQR_LOWER_THRESHOLD
    )
    iqr_limit_lower = tau_values <= iqr_lower_threshold

    adata.obs["tau_type"] = "normal"
    adata.obs.loc[(hard_limit | iqr_limit_upper), "tau_type"] = "high"
    adata.obs.loc[iqr_limit_lower, "tau_type"] = "low"
    adata.obs["tau"] = tau_values

    adata.uns["tau_thresholds"] = {}
    adata.uns["tau_thresholds"]["tau_upper_hard_limit"] = TAU_HARD_THRESHOLD
    adata.uns["tau_thresholds"]["tau_upper_iqr_limit"] = iqr_upper_threshold
    adata.uns["tau_thresholds"]["tau_lower_iqr_limit"] = iqr_lower_threshold

    total_nbr_aggregeates = np.sum(
        (adata.obs["tau_type"] == "high") | (adata.obs["tau_type"] == "low")
    )
    logger.info(
        ("Found %s aggregates. This is %1.0f%% of the components"),
        total_nbr_aggregeates,
        (total_nbr_aggregeates / len(adata.obs)) * 100,
    )

    return None if inplace else adata



"""Collapse molecules from a demultiplexed PNA dataset.

Copyright © 2024 Pixelgen Technologies AB.
"""

import dataclasses
import logging
import math
import multiprocessing
import time
import typing
from contextlib import contextmanager
from functools import cached_property
from pathlib import Path, PurePath
from typing import Literal

import joblib
import numpy as np
import numpy.typing as npt
import pandas as pd
import polars as pl
import pyarrow
import pyarrow as pa
import pyarrow.compute
import pyarrow.parquet
import pydantic

from pixelator.pna.collapse.adjacency import (
    build_network_cluster,
    build_network_directional,
)
from pixelator.pna.collapse.independent.shared_memory_registry import (
    ReadOnlySharedMemoryRegistry,
    SharedMemoryRegistry,
)
from pixelator.pna.collapse.utilities import (
    FAISSBackend,
    MoleculeCollapserAlgorithm,
    _collect_label_array_indices,
    _find_connected_components_cluster,
    _find_connected_components_directional,
    _split_chunks,
)
from pixelator.pna.config import PNAAntibodyPanel, PNAAssay
from pixelator.pna.demux.barcode_demuxer import PNAEmbedding
from pixelator.pna.report.models import SampleReport

logger = logging.getLogger("collapse")

CollapsibleRegion = Literal["umi-1", "umi-2"]


class MarkerCorrectionStats(pydantic.BaseModel):
    """Collect statistics of groups of molecules that will be collapsed.

    Attributes:
        marker: The name of the marker for which we are collapsing UMIs.
        region_id: The region id of the UMI to collapse. Either "umi-1" or "umi-2".
        input_reads: The total number of input reads to the collapse process.
        input_molecules: The total number of unique input reads to the collapse process.
        input_unique_umis: The total number of unique UMIs of `region_id` among the input reads.
        corrected_reads: The total number of reads that were modified by the error correction process.
        corrected_molecules: The total number of unique reads that were modified by the error correction process.
        corrected_unique_umis: The total number of unique UMIs of type `region_id` that were modified by the error correction process.
        output_unique_umis: The total number of unique UMIs of type `region_id` after correction.

    """

    marker: str
    region_id: CollapsibleRegion

    input_reads: int = 0
    input_molecules: int = 0
    input_unique_umis: int = 0

    corrected_reads: int = 0
    corrected_unique_umis: int = 0

    output_unique_umis: int = 0

    @pydantic.computed_field(return_type=int)  # type: ignore
    @property
    def output_reads(self) -> int:
        """The total number of reads after correction.

        This is always the same as `input_reads_count` but is included for consistency.
        """
        return self.input_reads

    @pydantic.computed_field(return_type=float)  # type: ignore
    @property
    def corrected_reads_fraction(self) -> float:
        """The fraction of input reads that were corrected."""
        return self.corrected_reads / self.input_reads if self.input_reads else 0.0

    @pydantic.computed_field(return_type=float)  # type: ignore
    @property
    def corrected_unique_umis_fraction(self) -> float:
        """The fraction of unique UMIs of `region_type` that were corrected."""
        return (
            self.corrected_unique_umis / self.input_unique_umis
            if self.input_unique_umis
            else 0.0
        )


@dataclasses.dataclass(slots=True)
class CollapseInputFile:
    """Keep track of the input file to collapse.

    :param path: Path to the input file.
    :param file_size: The total size of the input file.
    :param molecule_count: The number of rows in the input dataframe.
        i.e. The number of molecules (unique reads).
    """

    path: str
    file_size: int
    molecule_count: int


class SingleUMICollapseSampleReport(SampleReport):
    """Model for report data returned by the collapse stage."""

    report_type: str = "collapse-umi"

    region_id: str = pydantic.Field(
        ..., description="The UMI region that was collapsed."
    )

    input_reads: int = pydantic.Field(
        ..., description="The number of input reads processed."
    )

    input_molecules: int = pydantic.Field(
        ...,
        description="The number of molecules (unique reads) processed.",
    )

    input_unique_umis: int = pydantic.Field(
        ...,
        description="The number of unique input UMIs processed.",
    )

    output_unique_umis: int = pydantic.Field(
        ...,
        description="The number of unique UMIs after correction.",
    )

    corrected_unique_umis: int = pydantic.Field(
        ...,
        description="The number of UMIs with errors that were corrected.",
    )

    corrected_reads: int = pydantic.Field(
        ...,
        description="The sum of the reads over all UMIs that were corrected.",
    )

    processed_files: list[CollapseInputFile] = pydantic.Field(
        ..., description="The files processed during the collapse step."
    )

    markers: list[MarkerCorrectionStats] = pydantic.Field(
        ..., description="Correction statistics per UMI region for each marker group."
    )


T = typing.TypeVar("T")


class IndependentCollapseStatisticsCollector:
    """Collect statistics about the collapse process."""

    class _SummaryStatsDict(typing.TypedDict):
        input_reads: int
        input_molecules: int
        input_unique_umis: int
        corrected_reads: int
        corrected_unique_umis: int
        output_unique_umis: int

    def __init__(self, region_id: CollapsibleRegion) -> None:
        """Initialize the statistics collector."""
        self._processed_files: list[CollapseInputFile] = []
        self._marker_umi_data: dict[tuple[str, str], MarkerCorrectionStats] = {}
        self.region_id = region_id

    @property
    def markers(self) -> list[MarkerCorrectionStats]:
        """Return the per marker stats for the umi-1 region."""
        return [v for v in self._marker_umi_data.values()]

    @typing.overload
    def add_input_file(
        self, input_file: PurePath, molecule_count: int, file_size: int
    ) -> None: ...

    @typing.overload
    def add_input_file(
        self, input_file: Path, molecule_count: int, file_size: int | None = None
    ) -> None: ...

    def add_input_file(
        self,
        input_file: Path | PurePath,
        molecule_count: int,
        file_size: int | None = None,
    ) -> None:
        """Collect file statistics for an input file to the MoleculeCollapser.

        :param input_file: The input file to collapse.
        :param molecule_count: The number of molecules in the input file
        :raise TypeError: If file_size is not provided when input_file is a PurePath.
        """
        if file_size is None and isinstance(input_file, Path):
            file_size = input_file.stat(follow_symlinks=True).st_size
        elif file_size is None:
            raise TypeError("file_size must be provided if input_file is a PurePath")

        self._processed_files.append(
            CollapseInputFile(
                path=str(input_file),
                file_size=file_size,
                molecule_count=molecule_count,
            )
        )

    def add_marker_stats(
        self,
        stats: MarkerCorrectionStats,
    ) -> None:
        """Add statistics for a marker pair.

        Args:
            stats: The statistics for the correction of UMIs for reads with this marker.

        Raises:
            KeyError: If data for the marker pair already exists.

        """
        key = (stats.region_id, stats.marker)

        if stats.region_id != self.region_id:
            raise ValueError(
                f"Region id mismatch: {stats.region_id} != {self.region_id}"
            )

        if key in self._marker_umi_data:
            raise KeyError(
                f"Data for {stats.region_id} of {stats.marker} already exists"
            )

        self._marker_umi_data[key] = stats

    def get_combined_stats(self) -> _SummaryStatsDict:
        """Compute combined summary statistics from all the marker pairs."""
        df = pd.DataFrame([v.model_dump() for v in self._marker_umi_data.values()])

        aggregate_cols = [
            "input_reads",
            "input_molecules",
            "input_unique_umis",
            "corrected_reads",
            "corrected_unique_umis",
            "output_unique_umis",
        ]

        sums = df[aggregate_cols].sum()
        stats = sums.to_dict()
        return stats

    def to_sample_report(self, sample_id) -> SingleUMICollapseSampleReport:
        """Return the statistics as a SampleReport.

        Returns:
            A `SingleUMICollapseSampleReport` containing the statistics.

        """
        stats = self.get_combined_stats()
        return SingleUMICollapseSampleReport(
            sample_id=sample_id,
            product_id="single-cell-pna",
            region_id=self.region_id,
            **stats,
            processed_files=self._processed_files,
            markers=list(self._marker_umi_data.values()),
        )

    def to_dict(self) -> dict:
        """Return the statistics as a dictionary."""
        stats = self.get_combined_stats()
        return {
            "region_id": self.region_id,
            "processed_files": [dataclasses.asdict(f) for f in self._processed_files],
            "markers": [v.model_dump() for v in self._marker_umi_data.values()],
            **stats,
        }


@dataclasses.dataclass(slots=True, frozen=True)
class _DistributedResults:
    start: int
    stop: int

    #: A list of lists containing the molecule indices for each connected component
    unique_molecule_ids: list
    #: A list of the idx of the UMI that a cluster of UMIs need to be corrected to
    representative_indices: npt.ArrayLike
    corrected_reads_count: int

    collapsed_umi_count: int = 0
    output_umi_count: int = 0


class RegionCollapser:
    """Error correct UMI sequences based on similarity.

    Attributes:
        _umi1_data: A numpy array containing the unique UMI-1 sequences for the current processing batch.
            These are recoded to a 2-bit encoding and cast to a 64-bit integer.
        _umi2_data: A numpy array containing the unique UMI-2 sequences for the current processing batch.
            These are recoded to a 2-bit encoding and cast to a 64-bit integer.

        _db_to_molecule_idx: A numpy array containing for each of the input moleces the index of the unique UMI.
            This is used to link corrections of the unique umis back to all molecules that share the same UMI.
        _unique_umi_to_molecule_count: The number of molecules that map to each unique umi.
            This is used to map "unique umi" counts to the corresponding "molecule" counts.

    """

    _umi1_schema = pa.schema(
        fields=[
            pa.field("marker_1", pa.large_string()),
            pa.field("marker_2", pa.large_string()),
            pa.field("read_count", pa.uint16()),
            pa.field("original_umi1", pa.uint64()),
            pa.field("original_umi2", pa.uint64()),
            pa.field("uei", pa.uint64()),
            pa.field("corrected_umi1", pa.uint64()),
        ]
    )
    _umi2_schema = pa.schema(
        fields=[
            pa.field("marker_1", pa.large_string()),
            pa.field("marker_2", pa.large_string()),
            pa.field("read_count", pa.uint16()),
            pa.field("original_umi1", pa.uint64()),
            pa.field("original_umi2", pa.uint64()),
            pa.field("uei", pa.uint64()),
            pa.field("corrected_umi2", pa.uint64()),
        ]
    )

    def __init__(
        self,
        assay: PNAAssay,
        panel: PNAAntibodyPanel,
        region_id: CollapsibleRegion,
        max_mismatches: int | float = 0.1,
        algorithm: MoleculeCollapserAlgorithm = "directional",
        threads: int = -1,
        logger: logging.Logger | None = None,
        min_parallel_chunk_size: int = 500,
        similarity_backend: Literal["faiss"] = "faiss",
    ):
        """Initialize a RegionCollapser.

        Args:
            assay: The assay configuration.
            panel: The antibody panel configuration.
            region_id: The region id of the UMI to collapse. Either "umi-1" or "umi-2".
            max_mismatches: The maximum number of mismatches allowed when collapsing molecules.
                Either an integer >= 1 or a float in the range [0, 1).
            algorithm: The algorithm to use for collapsing molecules. Either "cluster" or "directional".
            threads: The number of threads to use for parallel processing.
            logger: The logger to use for output. The default is a logger named "collapse".
            min_parallel_chunk_size: The minimum number of com to process in parallel.
            similarity_backend: The backend to use for similarity search. Currently only "faiss".

        """
        self.assay = assay
        self.panel = panel
        self.algorithm = algorithm
        self.region_id = region_id

        if max_mismatches >= 1:
            if isinstance(max_mismatches, float):
                raise ValueError(
                    "max_mismatches must be either an integer value > 1"
                    " or a float value between 0 and 1."
                )
            self.max_mismatches = max_mismatches
        else:
            region_len = self.assay.get_region_by_id(self.region_id).max_len
            self.max_mismatches = math.ceil(region_len * max_mismatches)

        self.threads = threads
        self._memory = SharedMemoryRegistry()
        self._parallel_worker = joblib.Parallel(n_jobs=threads, return_as="list")
        self._logger = logger or logging.getLogger("collapse")
        self._simsearch = FAISSBackend(threads=threads)

        match self.region_id:
            case "umi-1":
                self._output_schema = self._umi1_schema
            case "umi-2":
                self._output_schema = self._umi2_schema
            case _:
                raise ValueError(f"Unknown region_id: {self.region_id}")

        self._embedding = PNAEmbedding(self.assay)
        self._stats = IndependentCollapseStatisticsCollector(region_id=self.region_id)
        self._marker_idx_to_name = np.vectorize(self.panel.markers.__getitem__)
        self.min_parallel_chunk_size = min_parallel_chunk_size

        # TODO vectorize these in the Embedding class itself

        self._umi1_region_extractor = np.vectorize(
            self._embedding.get_umi1_bytes, cache=True
        )
        self._umi2_region_extractor = np.vectorize(
            self._embedding.get_umi2_bytes, cache=True
        )
        self._uei_region_extractor = np.vectorize(
            self._embedding.get_uei_bytes, cache=True
        )
        self._compress_umi_embedding = np.vectorize(
            self._embedding.compress_umi_embedding, cache=True
        )
        self._compress_uei_embedding = np.vectorize(
            self._embedding.compress_uei_embedding, cache=True
        )

    def __enter__(self):
        """Enter the context manager for the RegionCollapser.

        This wraps the context managers for the shared memory manager, parallel worker pool,
        and the parquet writer.
        """
        self._memory = self._memory.__enter__()
        self._parallel_worker = self._parallel_worker.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit the context manager for the RegionCollapser.

        This wraps the context managers for the shared memory manager, parallel worker pool,
        and the parquet writer.
        """
        self._memory.__exit__(exc_type, exc_val, exc_tb)
        self._parallel_worker.__exit__(exc_type, exc_val, exc_tb)

    def statistics(self) -> IndependentCollapseStatisticsCollector:
        """Return the statistics collected during the collapse process."""
        return self._stats

    @cached_property
    def _effective_threads(self):
        if self.threads <= -1:
            return multiprocessing.cpu_count() + 1 - abs(self.threads)

        return self.threads

    @property
    def _umi_data(self) -> npt.NDArray:
        """Return either umi1 or umi2 data depending on the region_id."""
        return self._region_id_choice(self._umi1_data, self._umi2_data)

    @property
    def _db_to_molecule_idx(self) -> npt.NDArray:
        """Return the unique to molecule mapping for the current region_id."""
        return self._region_id_choice(
            self._db_to_molecule_idx_umi1, self._db_to_molecule_idx_umi2
        )

    def _extract_unique_umis(self, molecules: pl.Series):
        """Extract unique UMI sequences from a series of binary embeddings.

        Depending on the region_id either the UMI1 or UMI2 region is extracted.

        Args:
            molecules (pl.Series): A series containing the molecule data.

        Returns:
            A numpy array containing the unique UMI sequences.

        """
        umi1_data = self._umi1_region_extractor(molecules)
        unique_umi1s, inverse_umi1, counts_umi1 = np.unique(
            umi1_data, return_inverse=True, return_counts=True
        )
        del umi1_data

        umi2_data = self._umi2_region_extractor(molecules)
        unique_umi2s, inverse_umi2, counts_umi2 = np.unique(
            umi2_data, return_inverse=True, return_counts=True
        )
        del umi2_data

        # Re-encode the UMI sequences to a 32-bit integer.
        # These numeric embeddings will be writen to the output parquet files.
        # The original UMI embeddings are used to search for near-identical sequences since those are equidistant in the hamming space.
        self._umi1_data = self._compress_umi_embedding(unique_umi1s)
        self._umi2_data = self._compress_umi_embedding(unique_umi2s)

        # The indices in the unique umi array (eg: self._umi_data or the search db embedding)
        # that reconstruct the input molecule array.
        self._db_to_molecule_idx_umi1 = inverse_umi1
        self._db_to_molecule_idx_umi2 = inverse_umi2

        # The number of molecules that map to each unique umi.
        self._unique_umi_to_molecule_count_umi1 = counts_umi1

        # The number of molecules that map to each unique umi.
        self._unique_umi_to_molecule_count_umi2 = counts_umi2

        return (unique_umi1s, unique_umi2s)

    def _region_id_choice(self, umi1_res: T, umi2_res: T) -> T:
        """Return the first or second argument based on the current region id (umi1|umi2)."""
        # Avoid repeating this kind of logic everywhere
        if self.region_id == "umi-1":
            return umi1_res
        elif self.region_id == "umi-2":
            return umi2_res
        else:
            raise ValueError(f"Invalid region_id: {self.region_id}")

    @contextmanager
    def _init_shared_memory(self, data):
        """Return a contextmanager for the shared memory containing embeddings and read counts.

        This will allocate 2 shared memory buffers in the registry:

        db: A 2D array of uint8 with shape (num_unique_umis, 12) containing the unique umis
            Each row is a 12-byte binary vector representing a unique umi.

        read_counts: A 1D array of uint64 with shape (num_unique_umis) containing the read counts
        """
        # Inputs are 3 bit encoded UMI sequences
        # For the current 28 bases UMIs this results in 28/8 * 3 = 11 bytes per UMI
        # The UMI regions in the molecule embedding created by demux is zero-padded to 12 bytes.
        vector_length = 12

        unique_umi1s, unique_umi2s = self._extract_unique_umis(data["molecule"])
        unique_umis = self._region_id_choice(unique_umi1s, unique_umi2s)
        num_unique_umis = len(unique_umis)

        db = self._memory.allocate_array(
            "db", shape=(num_unique_umis, vector_length), dtype=np.uint8, zero_init=True
        )

        for idx, umi in enumerate(unique_umis):
            db[idx, 0 : len(umi)] = np.frombuffer(umi, dtype=np.uint8)

        read_counts = self._memory.allocate_array(
            "read_counts", shape=(num_unique_umis,), dtype=np.uint16, zero_init=False
        )

        # Sum the read counts for all molecules that map to the same unique umi.
        # These read_counts are than used as weights in the network based deduplication.
        read_counts[:] = np.bincount(
            self._db_to_molecule_idx,
            weights=data["read_count"],
            minlength=num_unique_umis,
        )

        yield db, read_counts

        self._memory.unlink_buffer("db")
        self._memory.unlink_buffer("read_counts")
        self._umi1_data = None
        self._umi2_data = None
        self._db_to_molecule_idx_umi1 = None
        self._db_to_molecule_idx_umi2 = None

    @cached_property
    def max_hamming_mismatches(self) -> int:
        """Return the maximum hamming distance for the number of allowed mismatches."""
        return self.max_mismatches * 2

    def process_file(self, path: Path, output: Path):
        """Collapse UMIs from a single parquet file containing molecular data."""
        logger = self._logger

        logger.info("Loading data from parquet: %s", str(path))
        df = pl.read_parquet(path)

        # Handle legacy data
        if "marker1" in df.columns or "marker2" in df.columns:
            df = df.rename({"marker1": "marker_1", "marker2": "marker_2"}, strict=False)

        self._stats.add_input_file(path, df.shape[0])

        logger.info("Partitioning dataset by marker pairs")

        partition_key = self._region_id_choice("marker_1", "marker_2")

        # Partition the data by the marker1 and marker2 columns and store each partition as a separate DataFrame
        partitions = df.partition_by(partition_key, as_dict=True, include_key=True)

        num_partitions = len(partitions)
        logger.info("Dataset partitioned into %s subsets", num_partitions)

        # Initialize the parquet writer
        writer = pyarrow.parquet.ParquetWriter(
            output, schema=self._output_schema, compression="zstd", compression_level=6
        )

        with writer as w:
            for idx, (group_by_keys, data) in enumerate(partitions.items()):
                # mypy cannot infer the type of group_by_keys
                marker_id = typing.cast(int, group_by_keys[0])
                stats = self._process_marker_group(
                    idx,
                    num_groups=num_partitions,
                    marker=marker_id,
                    data=data,
                    writer=w,
                )

                stats.region_id = self.region_id
                self._stats.add_marker_stats(stats)

    def _process_marker_group(
        self,
        idx: int,
        num_groups: int,
        marker: int,
        data: pl.DataFrame,
        writer: pa.parquet.ParquetWriter,
    ) -> MarkerCorrectionStats:
        """Process a group of reads from a single marker.

        :param idx: The index of the group in the partition
        :param num_groups: The total number of groups in the partition
        :param marker: The index of the marker in the panel
        :param data: The data for the group. A dataframe.
        :param writer: The parquet writer to stream output to.
        """
        starttime = time.time()
        logger = self._logger

        marker_name = self.panel.markers[marker]
        input_reads_count = int(data["read_count"].sum())

        local_stats = MarkerCorrectionStats(
            marker=marker_name,
            region_id=self.region_id,
            input_reads=input_reads_count,
            input_molecules=len(data),
        )

        logger.info(
            "Processing %s for %s reads from marker %s (%s/%s)",
            self.region_id,
            input_reads_count,
            marker_name,
            idx + 1,
            num_groups,
        )

        # Allocate shared memory for the binary vectors that make up the search space
        # and the read counts for each molecule. These are loaded with a context manager
        # to ensure proper cleanup
        with self._init_shared_memory(data) as shared_memory_data:
            db, read_counts = shared_memory_data

            local_stats.input_unique_umis = len(db)

            logger.info(
                "Building binary index of %s for %s from %s unique umis",
                self.region_id,
                marker_name,
                len(db),
            )

            # Build a binary index for given marker
            index = self._simsearch.build_index(db)

            logger.info(
                "Querying binary index of %s for %s for similar molecules",
                self.region_id,
                marker_name,
            )

            distances, indices = self._simsearch.search(index, db, min(16, len(db)))

            logger.info(
                "Building sparse adjacency matrix of %s for %s",
                self.region_id,
                marker_name,
            )

            if self.algorithm == "cluster":
                adjacency = build_network_cluster(
                    distances, indices, read_counts, self.max_hamming_mismatches
                )
            elif self.algorithm == "directional":
                adjacency = build_network_directional(
                    distances, indices, read_counts, self.max_hamming_mismatches
                )
            else:
                raise ValueError(f"Unknown algorithm: {self.algorithm}")

            logger.info(
                f"Collapsing clustered reads ({self.algorithm}) of %s for %s",
                self.region_id,
                marker_name,
            )

            corrected_umis, cluster_stats = self._process_molecule_graph(
                adjacency, marker_name=marker_name, local_stats=local_stats
            )

            logger.info(
                "Preparing collapsed output for %s from markers %s (%s/%s)",
                self.region_id,
                marker_name,
                idx + 1,
                num_groups,
            )

            # Get the original 2bit encoded UEI
            uei_bytes = self._uei_region_extractor(data["molecule"])
            uei_data = self._compress_uei_embedding(uei_bytes)
            # Get the original 2bit encoded UMIs
            original_umi1_data = self._umi1_data[self._db_to_molecule_idx_umi1]
            original_umi2_data = self._umi2_data[self._db_to_molecule_idx_umi2]

            new_data = data.with_columns(
                marker_1=self._marker_idx_to_name(data["marker_1"].to_numpy()),
                marker_2=self._marker_idx_to_name(data["marker_2"].to_numpy()),
                original_umi1=original_umi1_data,
                original_umi2=original_umi2_data,
                uei=uei_data,
            ).drop(["molecule"])

            table = new_data.to_arrow()

            corrected_name = self._region_id_choice("corrected_umi1", "corrected_umi2")
            table = table.add_column(
                6,
                self._output_schema.field(corrected_name),
                corrected_umis,
            )
            table = table.cast(self._output_schema)

            logger.info("Streaming %s records to parquet", len(table))
            writer.write_table(table)

            logger.info(
                "Completed processing %s for %s reads from marker %s (%s/%s)",
                self.region_id,
                input_reads_count,
                marker_name,
                idx + 1,
                num_groups,
            )

        return cluster_stats

    def _process_molecule_graph(
        self, csgraph, marker_name: str, local_stats: MarkerCorrectionStats
    ) -> tuple[pa.Array, MarkerCorrectionStats]:
        """Determine connected components and collapse the UMIs for each component.

        :param csgraph: The sparse adjacency matrix of the connected components
        :param local_stats: The statistics object for this marker pair
        :return: A tuple with a pyarrow Table with for each read the corrected UMI
        """
        _logger = self._logger

        n_components: int
        labels: npt.NDArray[np.int32]

        if self.algorithm == "cluster":
            n_components, labels = _find_connected_components_cluster(csgraph)
        elif self.algorithm == "directional":
            n_components, labels = _find_connected_components_directional(csgraph)
        else:
            raise ValueError(f"Unknown algorithm: {self.algorithm}")

        n_molecules = csgraph.shape[0]

        # Single pass over the labels to collect the indices of each connected component
        groups = _collect_label_array_indices(labels, n_components)

        memory_view = self._memory.read_only_view()
        embedding = self._embedding

        job_results: list[_DistributedResults]

        local_stats.input_unique_umis = len(self._umi_data)

        # Allocate output vectors in shared memory or as normal numpy arrays
        if self._effective_threads > 1 and n_components >= self.min_parallel_chunk_size:
            chunk_size = max(
                self.min_parallel_chunk_size, n_components // self._effective_threads
            )
            _logger.info(
                "Detecting unique %s from %s connected components for %s (parallel: [chunk_size=%s])",
                self.region_id,
                n_components,
                marker_name,
                chunk_size,
            )
            subranges = list(_split_chunks(n_components, chunk_size=chunk_size))

            job_results = self._parallel_worker(
                joblib.delayed(self._record_group_worker_fn)(
                    r, groups[slice(*r)], memory_view, embedding, n_molecules
                )
                for r in subranges
            )
        else:
            _logger.info(
                "Detecting unique %s from %s connected components for %s (serial)",
                self.region_id,
                n_components,
                marker_name,
            )

            job_results = [
                self._record_group_worker_fn(
                    (0, n_components), groups, memory_view, embedding, n_molecules
                )
            ]

        # An array of indices mapping the original unique reads into the corrected umi space
        # Initialize with the identity mapping
        corrected_umi_map = self._process_job_results(
            job_results, local_stats, n_molecules
        )

        _logger.info(
            "Corrected %s reads of %s (%.2f%%) or %s umis of %s unique umis (%.2f%%)",
            local_stats.corrected_reads,
            local_stats.input_reads,
            local_stats.corrected_reads_fraction * 100,
            local_stats.corrected_unique_umis,
            local_stats.input_unique_umis,
            local_stats.corrected_unique_umis_fraction * 100,
        )

        corrected_umis = self._build_corrected_umi_array(corrected_umi_map)

        return corrected_umis, local_stats

    def _process_job_results(
        self, job_results: list[_DistributedResults], local_stats, n_molecules: int
    ):
        """Process the results of the parallel job and build the corrected UMI map.

        Args:
            job_results: A list of _DistributedResults containing the results of the parallel job.
            local_stats: The statistics object for this marker pair.
            n_molecules: The number of molecules in the dataset.

        Returns:
            A numpy array with mapping input UMI indices to the corrected UMI.

        """
        # An array of indices mapping the original unique reads into the corrected umi space
        # Initialize with the identity mapping

        corrected_umi_map = np.zeros((n_molecules,), dtype=np.int64)

        for idx, r in enumerate(job_results):
            if r.start == r.stop:
                continue

            # Flatten the list of numpy indices that need to be corrected into a single array
            indices = np.concatenate(r.unique_molecule_ids, dtype=np.int64)

            # Repeat the unique umis for the number of umis that need to be corrected to that umi
            repeat_lengths = [len(x) for x in r.unique_molecule_ids]
            values = np.repeat(r.representative_indices, repeat_lengths)

            # Assign the values to the corrected_umi_map
            corrected_umi_map[indices] = values

            local_stats.corrected_reads += r.corrected_reads_count
            local_stats.corrected_unique_umis += r.collapsed_umi_count
            local_stats.output_unique_umis += r.output_umi_count

        return corrected_umi_map

    def _build_corrected_umi_array(
        self, corrected_umi_map: npt.NDArray[np.uint64]
    ) -> pyarrow.Array:
        """Build a table with the corrected UMI sequences.

        Args:
            corrected_umi_map: A numpy array containing the corrected UMI indices.

        Returns:
            A pyarrow Table with the original and corrected UMI encoded sequences.

        """
        # broadcast the corrected umi map to the original molecule indices using the reverse indices
        corrected_umi_indices = corrected_umi_map[self._db_to_molecule_idx]
        corrected_umis = self._umi_data[corrected_umi_indices]
        return pa.array(corrected_umis, type=pa.uint64())

    @staticmethod
    def _record_group_worker_fn(
        subrange: tuple[int, int],
        component_indices,
        memory: ReadOnlySharedMemoryRegistry,
        embedding: PNAEmbedding,
        n_molecules: int,
    ):
        """Process a batch of connected components.

        The database and read counts are loaded from shared memory to reduce
        python multiprocessing IPC overhead.

        :param subrange: The range of connected components to process.
            A tuple with the start and stop indices.
        :param component_indices: A list of lists containing the indices
            in the database and read counts vector for each connected component.
        :param db_shm: The shared memory buffer containing the binary vectors.
        :param read_counts_shm: The shared memory buffer containing the read counts.
        :param db_size: The size of the binary vectors memory buffer in bytes.
        :param embedding: The PNAEmbedding instance for encoding/decoding vectors.
        """
        db = memory.get_array("db")
        read_count = memory.get_array("read_counts")
        corrected_reads = 0
        corrected_umis = 0
        corrected_molecules = 0

        subrange_len = subrange[1] - subrange[0]
        unique_molecules_indices = []
        representative_indices: npt.NDArray[np.int64] = np.ndarray(
            shape=(subrange_len,), dtype=np.int64
        )

        for idx in range(subrange_len):
            this_cmp_indices = component_indices[idx]

            # Aggregate the read_count for this connected component
            molecule_read_counts = read_count[this_cmp_indices]
            component_reads = np.sum(molecule_read_counts)

            # Select the molecule with the most support as "representative" of the component
            # TODO: Voting on a representative molecule per nucleotide?
            representative_idx = np.argmax(molecule_read_counts)
            comp_idx = this_cmp_indices[representative_idx]
            corrected_reads += int(
                component_reads - molecule_read_counts[representative_idx]
            )

            representative_indices[idx] = comp_idx
            unique_molecules_indices.append(this_cmp_indices)
            corrected_umis += len(this_cmp_indices) - 1

        res = _DistributedResults(
            start=subrange[0],
            stop=subrange[1],
            representative_indices=representative_indices,
            unique_molecule_ids=unique_molecules_indices,
            corrected_reads_count=corrected_reads,
            collapsed_umi_count=corrected_umis,
            output_umi_count=len(unique_molecules_indices),
        )
        return res



"""Model for report data returned by the single-cell graph stage.

Copyright © 2023 Pixelgen Technologies AB.
"""

from __future__ import annotations

import pydantic

from pixelator.common.report.models.summary_statistics import SummaryStatistics
from pixelator.mpx.report.models.base import SampleReport


class GraphSampleReport(SampleReport):
    """Model for report data returned by the graph stage."""

    component_count: int = pydantic.Field(
        ...,
        description="The total number of components in the graph.",
    )

    molecule_count: int = pydantic.Field(
        ...,
        description="The total number of unique molecules in the graph.",
    )

    read_count: int = pydantic.Field(
        ...,
        description="The total number of reads in the graph.",
    )

    marker_count: int = pydantic.Field(
        ...,
        description="The total number of unique markers in the graph.",
    )

    read_count_per_molecule_stats: SummaryStatistics = pydantic.Field(
        ...,
        description="Summary statistics of the reads per component.",
    )

    a_pixel_count: int = pydantic.Field(
        ...,
        description="The number of unique A-pixels in the graph.",
    )

    b_pixel_count: int = pydantic.Field(
        ...,
        description="The number of unique B-pixels in the graph.",
    )

    fraction_molecules_in_largest_component: float = pydantic.Field(
        ...,
        ge=0,
        le=1,
        description="The fraction of all molecules that are located in the largest component.",
    )

    fraction_pixels_in_largest_component: float = pydantic.Field(
        ...,
        ge=0,
        le=1,
        description="The fraction of all pixels that are located in the largest component.",
    )

    edges_with_colliding_upi_count: int = pydantic.Field(
        ...,
        description="The number of edges with UPIs that have appeared both as UPIA and UPIB.",
    )

    edges_removed_in_multiplet_recovery_first_iteration: int = pydantic.Field(
        ...,
        description="The number of edges removed in the first iteration of multiplet recovery.",
    )

    edges_removed_in_multiplet_recovery_refinement: int = pydantic.Field(
        ...,
        description="The number of edges removed in the refinement of multiplet recovery.",
    )

    fraction_edges_removed_in_refinement: float = pydantic.Field(
        ...,
        ge=0,
        le=1,
        description="The fraction of total removed edges that are removed in the refinement of multiplet recovery.",
    )

    @pydantic.computed_field(
        return_type=float,
        description="The ratio of the total number of A-pixels and the total number of B-pixels in the graph.",
    )
    def a_pixel_b_pixel_ratio(self):  # noqa: D102
        return self.a_pixel_count / self.b_pixel_count

    @pydantic.computed_field(  # type: ignore
        return_type=int,
        description="The total number of unique pixels in the graph.",
    )
    @property
    def pixel_count(self) -> int:  # noqa: D102
        return self.a_pixel_count + self.b_pixel_count



"""Copyright © 2023 Pixelgen Technologies AB."""

from __future__ import annotations

import enum
import functools
import itertools
import logging
import typing
from collections import defaultdict
from pathlib import Path, PurePath
from typing import Dict, List, Optional, Tuple

from pixelator.common.types import PathType
from pixelator.common.utils import flatten, get_sample_name

logger = logging.getLogger(__name__)


class SingleCellStage(enum.Enum):
    """Enum for the different stages of the single-cell pipeline."""

    AMPLICON = "amplicon"
    PREQC = "preqc"
    ADAPTERQC = "adapterqc"
    DEMUX = "demux"
    COLLAPSE = "collapse"
    GRAPH = "graph"
    ANNOTATE = "annotate"
    LAYOUT = "layout"
    ANALYSIS = "analysis"
    REPORT = "report"


SingleCellStageLiteral = typing.Literal[
    "amplicon",
    "preqc",
    "adapterqc",
    "demux",
    "collapse",
    "graph",
    "annotate",
    "layout",
    "analysis",
    "report",
]

# Duplicating the keys is unavoidable here
WorkdirCacheKey: typing.TypeAlias = typing.Literal[
    "metadata",
    "single-cell-mpx amplicon",
    "single-cell-mpx preqc",
    "single-cell-mpx adapterqc",
    "single-cell-mpx demux",
    "single-cell-mpx collapse",
    "single-cell-mpx graph",
    "single-cell-mpx annotate",
    "single-cell-mpx annotate dataset",
    "single-cell-mpx annotate raw_components_metrics",
    "single-cell-mpx layout",
    "single-cell-mpx analysis",
    "single-cell-mpx report",
]

SINGLE_CELL_STAGES_TO_CACHE_KEY_MAPPING: dict[
    SingleCellStageLiteral, WorkdirCacheKey
] = {
    "amplicon": "single-cell-mpx amplicon",
    "preqc": "single-cell-mpx preqc",
    "adapterqc": "single-cell-mpx adapterqc",
    "demux": "single-cell-mpx demux",
    "collapse": "single-cell-mpx collapse",
    "graph": "single-cell-mpx graph",
    "annotate": "single-cell-mpx annotate",
    "layout": "single-cell-mpx layout",
    "analysis": "single-cell-mpx analysis",
    "report": "single-cell-mpx report",
}


CacheDictType = dict[WorkdirCacheKey, dict[str, Path] | dict[str, list[Path]]]


class WorkdirOutputNotFound(Exception):
    """Raised when an output file is is not found for a specific sample."""

    def __init__(
        self,
        *args,
        message: str | None = None,
        sample_id: str | None = None,
        pattern: str | None = None,
        workdir: Path | None = None,
    ) -> None:
        """Initialize the exception.

        :param args: Positional arguments to pass to the base
        :param message: A custom message to use
        :param sample_id: The sample id for which the output was not found
        :param pattern: The pattern used to search for the output
        :param workdir: The workdir folder where the output was searched for
        """
        super().__init__(*args)
        self.sample_id = sample_id
        self.pattern = pattern
        self.workdir = workdir
        self.message = message or (
            f'No output files found for sample "{self.sample_id}":\n'
            f'  using pattern "{self.pattern}" in "{self.workdir}"'
        )

    def __str__(self):
        """Return a string representation of the exception."""
        return self.message


class PixelatorWorkdir:
    """Tools to manage and collect files from a pixelator workdir folder."""

    # A dict that maps a key (usually a key per subcommand) to a glob pattern
    # that will be used to search for report files in the workdir folder.
    _SEARCH_PATTERNS: Dict[WorkdirCacheKey, str] = {
        "metadata": "**/*.meta.json",
        "single-cell-mpx amplicon": "amplicon/*.report.json",
        "single-cell-mpx preqc": "preqc/*.report.json",
        "single-cell-mpx adapterqc": "adapterqc/*.report.json",
        "single-cell-mpx demux": "demux/*.report.json",
        "single-cell-mpx collapse": "collapse/*.report.json",
        "single-cell-mpx graph": "graph/*.report.json",
        "single-cell-mpx annotate": "annotate/*.report.json",
        "single-cell-mpx annotate dataset": "annotate/*.dataset.pxl",
        "single-cell-mpx annotate raw_components_metrics": "annotate/*.raw_components_metrics.csv.gz",
        "single-cell-mpx layout": "layout/*.report.json",
        "single-cell-mpx analysis": "analysis/*.report.json",
        "single-cell-mpx report": "report/*.report.json",
    }

    _SEARCH_ANNOTATE_DATASET = "annotate/*.dataset.pxl"

    def __init__(self, basedir: PathType):
        """Initialize the PixelatorWorkdir object."""
        self.basedir = Path(basedir)
        self._cache: CacheDictType = {}
        self._sample_ids: set[str] | None = None

    def scan(self):
        """Retrieve all reports and metadata from the workdir folder."""
        self.metadata_files()
        for stage in SingleCellStage:
            # Some stages may not be present in the workdir, but
            # we still collect whatever is found
            try:
                self.single_cell_report(stage, cache=False)
            except FileNotFoundError:
                continue

    def stage_dir(self, stage: SingleCellStage | SingleCellStageLiteral):
        """Create a subcommand directory.

        :param stage: The subcommand to create a directory for
        :return: The absolute path to the subcommand directory
        """
        dirname = stage.value if isinstance(stage, enum.Enum) else stage
        subcommand_path = self.basedir / dirname
        subcommand_path.mkdir(exist_ok=True, parents=True)
        return subcommand_path

    def _check_folder(self, name: str) -> Path:
        """Check that the directory for a specific subcommand is present.

        :param name: The name of the subcommand folder
        :returns: The absolute path to the subcommand folder
        :raises FileNotFoundError: If the folder is missing
        """
        source_path = Path(self.basedir) / name
        if not source_path.is_dir():
            raise FileNotFoundError(f"{name} folder missing in {self.basedir}")

        return source_path.absolute()

    def samples(self, cache: bool = True) -> set[str]:
        """Return a list of all sample_ids encountered in the current workdir."""
        if cache and self._sample_ids is not None:
            return self._sample_ids

        self.scan()
        sids: set[str] = set()
        for pattern, data in self._cache.items():
            sids.update(data.keys())

        self._sample_ids = sids
        return self._sample_ids

    def _check_pattern_basedir(self, pattern_id: WorkdirCacheKey) -> Path:
        pattern = PurePath(self._SEARCH_PATTERNS[pattern_id])
        basedir = pattern.parents[0]
        return self._check_folder(str(basedir))

    def _collect_files_by_pattern(
        self, pattern: str, prefix: Optional[Path] = None
    ) -> List[Tuple[str, List[Path]]]:
        """Collect a set of report files from a subcommand folder.

        :param pattern: A glob pattern to search for.
        :param prefix: A prefix path to search in relative to the workdir
        :return: A list of tuples with sample name and a list of report files
        """
        # collect the metrics files
        search_dir = self.basedir / prefix if prefix else self.basedir

        files = list(search_dir.rglob(pattern))

        tuples = []
        for f in files:
            sample_name = get_sample_name(f)
            tuples.append((sample_name, f))

        tuples.sort(key=lambda x: x[0])
        grouped_files = [
            (str(sample), [p for (s, p) in grouper])
            for sample, grouper in itertools.groupby(tuples, key=lambda x: x[0])
        ]
        return grouped_files

    def _collect_files_by_key(
        self, search_key: WorkdirCacheKey, prefix: Optional[Path] = None
    ) -> List[Tuple[str, List[Path]]]:
        """Collect a set of report files from a subcommand folder.

        :param search_key: The key of an entry in the `ivar:_SEARCH_PATTERNS` dict
        :param prefix: A prefix path to search in relative to the workdir
        :return: A list of tuples with sample name and a list of report files
        """
        # collect the metrics files
        pattern = self._SEARCH_PATTERNS[search_key]
        return self._collect_files_by_pattern(pattern, prefix)

    def _collect_metadata_files(self) -> dict[str, list[Path]]:
        """Collect all metadata files.

        Metadata files are those that end with `.meta.json`.
        These typically contains information about the pixelator version
        and the commandline line arguments used to run a specific subcommand.
        The files will be returned as a flat list

        :raises AssertionError: If no metadata files are found
        :return: A list of metadata files
        """
        files = self._collect_files_by_key("metadata")
        if files is None or len(files) == 0:
            logging.warning(f"No metadata files found in {self.basedir}")
            return {}

        flat_files = list(
            itertools.chain.from_iterable(paths for sample, paths in files)
        )

        metadata_files_index = defaultdict(list)

        for file in flat_files:
            metadata_files_index[get_sample_name(file.name)].append(file)

        return {**metadata_files_index}

    def _collect_output_files(self, pattern_id: WorkdirCacheKey) -> dict[str, Path]:
        """Collect output files from a pixelator workdir.

        The pattern id is used to fetch a glob pattern from py::attr:`_SEARCH_PATTERNS`
        this glob is relative to the workdir folder.

        :param pattern_id: The pattern id to use for searching
        :return: A dictionary with sample names as keys and the output file as value
        :raises AssertionError: If there is more than 1 match with the pattern
        """
        self._check_pattern_basedir(pattern_id)
        collected_files = self._collect_files_by_key(pattern_id)
        stage_reports_index = {}

        for sample, filelist in collected_files:
            if len(filelist) != 1:
                raise AssertionError(
                    f"Expected exactly one report file for sample {sample}"
                )
            stage_reports_index[sample] = filelist[0]

        return stage_reports_index

    def metadata_files(
        self, sample: Optional[str] = None, cache: bool = True
    ) -> list[Path]:
        """Return a list of metadata files.

        :param sample: The sample name to filter on
        :param cache: Whether to return cached result if available
        :return: A list of metadata files
        """
        cache_key: WorkdirCacheKey = "metadata"

        def func() -> dict[str, list[Path]]:
            report_per_sample = self._collect_metadata_files()

            out: dict[str, list[Path]] = defaultdict(list)
            for k, v in report_per_sample.items():
                if isinstance(v, list):
                    out[k].extend(v)
                else:
                    out[k].append(v)

            return out

        return self._cached_reports_implementation(cache_key, func, sample, cache)

    @typing.overload
    def _cached_reports_implementation(
        self,
        cache_key: WorkdirCacheKey,
        fn: typing.Callable[[], dict[str, list[Path]]],
        sample: Optional[str] = None,
        cache: bool = True,
    ) -> list[Path]: ...

    @typing.overload
    def _cached_reports_implementation(
        self,
        cache_key: WorkdirCacheKey,
        fn: typing.Callable[[], dict[str, Path]],
        sample: Optional[str] = None,
        cache: bool = True,
    ) -> Path: ...

    def _cached_reports_implementation(
        self,
        cache_key: WorkdirCacheKey,
        fn: typing.Callable[[], dict[str, Path] | dict[str, list[Path]]],
        sample: Optional[str] = None,
        cache: bool = True,
    ) -> Path | list[Path]:
        if not cache or self._cache.get(cache_key) is None:
            self._cache[cache_key] = fn()

        if sample:
            res = self._cache[cache_key].get(sample)
            if res is None:
                raise WorkdirOutputNotFound(
                    sample_id=sample,
                    pattern=self._SEARCH_PATTERNS[cache_key],
                    workdir=self.basedir,
                )
            return res

        return list(flatten(self._cache[cache_key].values()))

    @typing.overload
    def single_cell_report(
        self,
        stage: SingleCellStage | SingleCellStageLiteral,
        sample: None = None,
        *,
        cache: bool = True,
    ) -> List[Path]: ...

    @typing.overload
    def single_cell_report(
        self,
        stage: SingleCellStage | SingleCellStageLiteral,
        sample: str,
        *,
        cache: bool = True,
    ) -> Path: ...

    def single_cell_report(
        self,
        stage: SingleCellStage | SingleCellStageLiteral,
        sample: Optional[str] = None,
        *,
        cache: bool = True,
    ) -> Path | List[Path]:
        """Retrieve a report from a stage of the single-cell pipeline.

        :param stage: The name of the stage to retrieve a report from
        :param sample: The sample name to filter on
        :param cache: Whether to return cached result if available
        :return: A list of metadata files
        """
        stage_key = stage.value if isinstance(stage, enum.Enum) else stage
        cache_key: WorkdirCacheKey = SINGLE_CELL_STAGES_TO_CACHE_KEY_MAPPING[stage_key]
        func = functools.partial(self._collect_output_files, cache_key)
        return self._cached_reports_implementation(cache_key, func, sample, cache)

    @typing.overload
    def filtered_dataset(
        self,
        sample: None = None,
        cache: bool = True,
    ) -> list[Path]: ...

    @typing.overload
    def filtered_dataset(
        self,
        sample: str,
        cache: bool = True,
    ) -> Path: ...

    def filtered_dataset(
        self, sample: str | None = None, cache: bool = True
    ) -> Path | list[Path]:
        """Return the path to a filtered dataset for a sample.

        This is the output `dataset.pxl` file from single-cell annotate

        :param sample: The sample name to retrieve the filtered dataset for
        :param cache: Whether to return cached result if available
        :return: The path to the filtered dataset if a sample is given, otherwise a list of all datasets
        """
        cache_key: WorkdirCacheKey = "single-cell-mpx annotate dataset"
        func = functools.partial(self._collect_output_files, cache_key)
        return self._cached_reports_implementation(cache_key, func, sample, cache)

    @typing.overload
    def raw_component_metrics(
        self,
        sample: None = None,
        cache: bool = True,
    ) -> list[Path]: ...

    @typing.overload
    def raw_component_metrics(
        self,
        sample: str,
        cache: bool = True,
    ) -> Path: ...

    def raw_component_metrics(
        self, sample: Optional[str] = None, cache: bool = True
    ) -> Path | list[Path]:
        """Return the path to a raw_component_metrics file.

        This is generated by single-cell annotate

        :param sample: The sample name to retrieve the filtered dataset for
        :param cache: Whether to return cached results if available
        :return: The path to a raw_component_metrics file if a sample is given, otherwise a list of all files
        """
        cache_key: WorkdirCacheKey = "single-cell-mpx annotate raw_components_metrics"
        func = functools.partial(self._collect_output_files, cache_key)
        return self._cached_reports_implementation(cache_key, func, sample, cache)



"""Top-level package for Pixelator.

Copyright © 2022 Pixelgen Technologies AB.
"""

from importlib import metadata

__version__ = "0.22.1"

try:
    __version__ = metadata.version("pixelgen-pixelator")
except metadata.PackageNotFoundError:
    pass


# Adding imports here as shortcuts to be able to import like
# import pixelator as mpx
# mpx.read("<file path>")
# and similar
from pixelator.mpx.pixeldataset import PixelDataset as MPXPixelDataset  # noqa
from pixelator.mpx.pixeldataset import read as read_mpx
from pixelator.mpx.pixeldataset.aggregation import (
    simple_aggregate as simple_aggregate_mpx,
)  # noqa

from pixelator.pna import read as read_pna

__all__ = ["read_mpx", "simple_aggregate_mpx", "MPXPixelDataset", "read_pna"]
